import numpy as np

from credsweeper.common.constants import ML_HUNK
from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInPostamble(WordIn):
    """Feature is true if line contains at least one word from predefined list."""

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns true if any words in a part of line after value"""
        postamble_end = len(candidate.line_data_list[0].line) \
            if len(candidate.line_data_list[0].line) < candidate.line_data_list[0].value_end + ML_HUNK \
            else candidate.line_data_list[0].value_end + ML_HUNK
        postamble = candidate.line_data_list[0].line[candidate.line_data_list[0].value_end:postamble_end].strip()

        return self.word_in_(postamble.lower()) if postamble else np.array([self.zero])



import logging
from abc import ABC
from typing import List, Optional, Tuple

from credsweeper.common.constants import MIN_DATA_LEN
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class StringsScanner(AbstractScanner, ABC):
    """Implements known binary file scanning with ASCII strings representations"""

    @staticmethod
    def get_enumerated_lines(data: bytes) -> List[Tuple[int, str]]:
        """Processes binary to found ASCII strings. Use offset instead line number."""
        enumerated_lines = []
        offset = -1
        line_items = []
        for n, x in enumerate(data):
            if 0x09 == x or 0x20 <= x <= 0x7E:
                # TAB, SPACE and visible ASCII symbols
                if 0 > offset:
                    # use start of string as line number
                    offset = n
                line_items.append(chr(x))
                continue
            if MIN_DATA_LEN <= len(line_items):
                # add valuable lines only
                enumerated_lines.append((offset, ''.join(line_items)))
            offset = -1
            line_items.clear()
        if MIN_DATA_LEN <= len(line_items):
            enumerated_lines.append((offset, ''.join(line_items)))
        return enumerated_lines

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Scan binary files for ASCII strings"""
        if strings := StringsScanner.get_enumerated_lines(data_provider.data):
            string_data_provider = StringContentProvider(lines=[x[1] for x in strings],
                                                         line_numbers=[x[0] for x in strings],
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|STRINGS")
            return self.scanner.scan(string_data_provider)
        return None if strings is None else []



import ast
import base64
import contextlib
import json
import logging
import math
import os
import random
import re
import string
from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional, Union

import numpy as np
import yaml
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.dh import DHPrivateKey, DHPublicKey
from cryptography.hazmat.primitives.asymmetric.dsa import DSAPrivateKey, DSAPublicKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey, EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey, Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PublicKey, X25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.x448 import X448PublicKey, X448PrivateKey
from cryptography.hazmat.primitives.serialization import load_der_private_key
from cryptography.hazmat.primitives.serialization.pkcs12 import load_key_and_certificates
from lxml import etree

from credsweeper.common.constants import AVAILABLE_ENCODINGS, \
    DEFAULT_ENCODING, LATIN_1, CHUNK_SIZE, MAX_LINE_LENGTH, CHUNK_STEP_SIZE, ASCII

logger = logging.getLogger(__name__)


class Util:
    """Class that contains different useful methods."""

    @staticmethod
    def get_extension(file_path: str, lower=True) -> str:
        """Return extension of file in lower case by default e.g.: '.txt', '.JPG'"""
        _, extension = os.path.splitext(str(file_path))
        return extension.lower() if lower else extension

    @staticmethod
    def get_regex_combine_or(re_strs: List[str]) -> str:
        """Routine combination for regex 'or'"""
        result = "(?:"

        for elem in re_strs:
            result += elem + "|"

        if result[-1] == "|":
            result = result[:-1]
        result += ")"

        return result

    @staticmethod
    def get_shannon_entropy(data: Union[str, bytes]) -> float:
        """Borrowed from http://blog.dkbza.org/2007/05/scanning-data-for-entropy-anomalies.html."""
        if not data:
            return 0.0
        size = len(data)
        _uniq, counts = np.unique(list(data), return_counts=True)
        probabilities = counts / size
        entropy = -float(np.sum(probabilities * np.log2(probabilities)))
        return entropy

    # Precalculated data for speedup
    MIN_DATA_ENTROPY: Dict[int, float] = {
        16: 1.66973671780348,
        20: 2.07723544540831,
        32: 3.25392803184602,
        40: 3.64853567064867,
        64: 4.57756933688035,
        384: 7.39,
        512: 7.55,
    }

    @staticmethod
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal entropy for size of random data. Precalculated data is applied for speedup"""
        if x in Util.MIN_DATA_ENTROPY:
            y = Util.MIN_DATA_ENTROPY[x]
        elif 8 < x < 64:
            # approximated for range 12 - 64
            _x = x - 8
            y = ((0.000016617804 * _x - 0.002695077) * _x + 0.170393) * _x + 0.4
        elif 64 < x < 384:
            # logarithm base 2 - slow, but precise
            _x = x - 8
            y = 1.095884 * math.log2(_x) - 1.90156
        elif 384 < x < 512:
            # solved for 384 - 512
            y = -0.11215851 * math.log2(x)**2 + 2.34303484 * math.log2(x) - 4.4466237
        else:
            # less or equal to 8 bytes might have 0 entropy
            y = 0
        return y

    @staticmethod
    def is_ascii_entropy_validate(data: bytes) -> bool:
        """
        Tests small data sequence (<256) for data randomness by testing for ascii and shannon entropy
        Returns True when data is an ASCII symbols or have small entropy
        """
        if not data:
            return True
        data_len = len(data)
        if 9 > data_len:
            # even random data may have 0 entropy for length of 8 bytes and less
            return True
        entropy = 0.
        cells = [int(0)] * 256
        ascii_test = True
        # "basket" sorting approach
        for x in data:
            cells[x] += 1
            if ascii_test and 0b10000000 & x:
                ascii_test = False
        if ascii_test:
            # only ascii symbols found
            return True
        left = 0.
        step = 256.0 / data_len
        right = left + step
        while left < 256:
            cell_sum = 0
            i = int(left)
            r = int(right)
            while i < r and i < 256:
                cell_sum += cells[i]
                i += 1
            p_x = float(cell_sum) / data_len
            if p_x > 0:
                entropy += -p_x * math.log2(p_x)
            left = right
            right += step
        min_entropy = Util.get_min_data_entropy(data_len)
        return entropy < min_entropy

    @staticmethod
    def is_binary(data: Union[bytes, bytearray]) -> bool:
        """
        Returns True when two zeroes sequence is found in begin of data.
        The sequence never exists in text format (UTF-8, UTF-16). UTF-32 is not supported.
        """
        if isinstance(data, (bytes, bytearray)) and 0 <= data.find(b"\0\0", 0, MAX_LINE_LENGTH):
            return True
        return False

    NOT_LATIN1_PRINTABLE_SET = set(range(0, 256)) \
        .difference(set(x for x in string.printable.encode(ASCII))) \
        .difference({0x1B}) \
        .difference(set(x for x in range(0xA0, 0x100)))

    @staticmethod
    def is_latin1(data: Union[bytes, bytearray]) -> bool:
        """Returns True when data looks like LATIN-1 for first MAX_LINE_LENGTH bytes."""
        result = False
        if data:
            non_latin1_cnt = sum(1 for x in data[:MAX_LINE_LENGTH] if x in Util.NOT_LATIN1_PRINTABLE_SET)
            # experiment for 255217 binary files shown avg = 0.268264 ± 0.168767, so let choose minimal
            chunk_len = min(MAX_LINE_LENGTH, len(data))
            result = bool(0.1 > non_latin1_cnt / chunk_len)
        return result

    @staticmethod
    def read_file(path: Union[str, Path], encodings: Optional[List[str]] = None) -> List[str]:
        """Read the file content using different encodings.

        Try to read the contents of the file according to the list of encodings "encodings" as soon as reading
        occurs without any exceptions, the data is returned in the current encoding

        Args:
            path: path to file
            encodings: supported encodings

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned

        """
        data = Util.read_data(path)
        return Util.decode_bytes(data, encodings)

    @staticmethod
    def decode_text(content: bytes, encodings: Optional[List[str]] = None) -> Optional[str]:
        """Decode content using different encodings.

        Try to decode bytes according to the list of encodings "encodings"
        occurs without any exceptions. UTF-16 requires BOM

        Args:
            content: raw data that might be text
            encodings: supported encodings

        Return:
            Decoded text in str for any suitable encoding
            or None when binary data detected

        """
        text = None
        binary_suggest = False
        if encodings is None:
            encodings = AVAILABLE_ENCODINGS
        for encoding in encodings:
            try:
                if binary_suggest and LATIN_1 == encoding and (Util.is_binary(content) or not Util.is_latin1(content)):
                    # LATIN_1 may convert data (bytes in range 0x80:0xFF are transformed)
                    break
                _text = content.decode(encoding=encoding, errors="strict")
                if content != _text.encode(encoding=encoding, errors="strict"):
                    # the check helps to detect a real encoding
                    raise UnicodeError
                text = _text
                break
            except UnicodeError:
                binary_suggest = True
                logger.info(f"UnicodeError: Can't decode content as {encoding}.")
            except Exception as exc:
                logger.error(f"Unexpected Error: Can't read content as {encoding}. Error message: {exc}")
        return text

    @staticmethod
    def split_text(text: str) -> List[str]:
        """Splits a text into lines, handling all common line endings (e.g., LF, CRLF, CR)."""
        return text.replace("\r\n", '\n').replace('\r', '\n').split('\n')

    @staticmethod
    def decode_bytes(content: bytes, encodings: Optional[List[str]] = None) -> List[str]:
        """Decode content using different encodings.

        Try to decode bytes according to the list of encodings "encodings"
        occurs without any exceptions. UTF-16 requires BOM

        Args:
            content: raw data that might be text
            encodings: supported encodings

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned
            Also empty list will be returned after last encoding and 0 symbol is present in lines not at end

        """
        if text := Util.decode_text(content, encodings):
            lines = Util.split_text(text)
        else:
            lines = []
        return lines

    @staticmethod
    def get_asn1_size(data: Union[bytes, bytearray]) -> int:
        """Only sequence type 0x30 and size correctness are checked
        Returns size of ASN1 data over 128 bytes or 0 if no interested data
        """
        if isinstance(data, (bytes, bytearray)) and 2 <= len(data) and 0x30 == data[0]:
            # https://www.oss.com/asn1/resources/asn1-made-simple/asn1-quick-reference/basic-encoding-rules.html#Lengths
            length = data[1]
            if 0x80 == length:
                if data.endswith(b"\x00\x00"):
                    # assume, all data are ASN1 of various size
                    return len(data)
                else:
                    # skip the case where the ASN1 size is smaller than the actual data
                    return 0
            elif 0x80 < length:
                byte_len = 0x7F & length
                len_limit = 2 + byte_len
                if 4 >= byte_len and len(data) >= len_limit:
                    length = 0
                    for i in range(2, len_limit):
                        length <<= 8
                        length |= data[i]
                    if len(data) >= length + len_limit:
                        return length + len_limit
                else:
                    # unsupported huge size
                    return 0
            else:
                # less than 0x80
                if len(data) >= length + 2:
                    return length + 2
        return 0

    @staticmethod
    def read_data(path: Union[str, Path]) -> Optional[bytes]:
        """Read the file bytes as is.

        Try to read the data of the file.

        Args:
            path: path to file

        Return:
            list of file rows in a suitable encoding from "encodings",
            if none of the encodings match, an empty list will be returned

        """

        try:
            with open(path, "rb") as file:
                return file.read()
        except Exception as exc:
            logger.error(f"Unexpected Error: Can not read '{path}'. Error message: '{exc}'")
        return None

    @staticmethod
    def get_xml_from_lines(xml_lines: List[str]) -> Tuple[Optional[List[str]], Optional[List[int]]]:
        """Parse xml data from list of string and return List of str.

        Args:
            xml_lines: list of lines of xml data

        Return:
            List of formatted string(f"{root.tag} : {root.text}")

        Raises:
            xml exception

        """
        lines = []
        line_nums = []
        tree = etree.fromstringlist(xml_lines)
        for element in tree.iter():
            tag = Util.extract_element_data(element, "tag")
            text = Util.extract_element_data(element, "text")
            lines.append(f"{tag} : {text}")
            line_nums.append(element.sourceline)
        return lines, line_nums

    @staticmethod
    def extract_element_data(element: Any, attr: str) -> str:
        """Extract xml element data to string.

        Try to extract the xml data and strip() the string.

        Args:
            element: xml element
            attr: attribute name

        Return:
            String xml data with strip()

        """
        element_attr: Any = getattr(element, attr)
        if element_attr is None or not isinstance(element_attr, str):
            return ''
        return str(element_attr).strip()

    @staticmethod
    def json_load(file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> Any:
        """Load dictionary from JSON file"""
        try:
            with open(file_path, "r", encoding=encoding) as f:
                return json.load(f)
        except Exception as exc:
            logging.error(f"Failed to read: {file_path} {exc}")
        return None

    @staticmethod
    def json_dump(obj: Any, file_path: Union[str, Path], encoding=DEFAULT_ENCODING, indent=4) -> None:
        """Write dictionary to JSON file"""
        try:
            with open(file_path, "w", encoding=encoding) as f:
                json.dump(obj, f, indent=indent)
        except Exception as exc:
            logging.error(f"Failed to write: {file_path} {exc}")

    @staticmethod
    def yaml_load(file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> Any:
        """Load dictionary from YAML file"""
        try:
            with open(file_path, "r", encoding=encoding) as f:
                return yaml.safe_load(f)
        except Exception as exc:
            logger.error(f"Failed to read {file_path} {exc}")
        return None

    @staticmethod
    def yaml_dump(obj: Any, file_path: Union[str, Path], encoding=DEFAULT_ENCODING) -> None:
        """Write dictionary to YAML file"""
        try:
            with open(file_path, "w", encoding=encoding) as f:
                yaml.dump(obj, f)
        except Exception as exc:
            logging.error(f"Failed to write: {file_path} {exc}")

    @staticmethod
    def parse_python(source: str) -> List[Any]:
        """Parse python source and back to remove strings merge and line wrap"""
        src = ast.parse(source)
        result = ast.unparse(src).splitlines()
        return result

    PEM_CLEANING_PATTERN = re.compile(r"\\[tnrvf]")
    WHITESPACE_TRANS_TABLE = str.maketrans('', '', string.whitespace)

    @staticmethod
    def decode_base64(text: str, padding_safe: bool = False, urlsafe_detect=False) -> bytes:
        """decode text to bytes with / without padding detect and urlsafe symbols"""
        value = text.translate(Util.WHITESPACE_TRANS_TABLE)
        if padding_safe:
            value = value.rstrip('=')  # python 3.10 workaround
            pad_num = 0x3 & len(value)
            if pad_num:
                value += '=' * (4 - pad_num)
        if urlsafe_detect and ('-' in value or '_' in value):
            decoded = base64.b64decode(value, altchars=b"-_", validate=True)
        else:
            decoded = base64.b64decode(value, validate=True)
        return decoded

    @staticmethod
    def load_pk(data: bytes, password: Optional[bytes] = None) -> Optional[PrivateKeyTypes]:
        """Try to load private key from PKCS1, PKCS8 and PKCS12 formats"""
        with contextlib.suppress(Exception):
            # PKCS1, PKCS8 probes
            private_key = load_der_private_key(data, password)
            return private_key
        with contextlib.suppress(Exception):
            # PKCS12 probe
            private_key, _certificate, _additional_certificates = load_key_and_certificates(data, password)
            return private_key
        return None

    RANDOM_DATA = random.randbytes(20)

    @staticmethod
    def check_pk(pkey: PrivateKeyTypes) -> bool:
        """Check private key with encrypt-decrypt random data"""
        if isinstance(pkey, (EllipticCurvePrivateKey, DSAPrivateKey, Ed448PrivateKey, Ed25519PrivateKey, DHPrivateKey,
                             X448PrivateKey, X25519PrivateKey)):
            # One does not simply perform check the keys
            return True
        if isinstance(pkey, (EllipticCurvePublicKey, DSAPublicKey, Ed448PublicKey, Ed25519PublicKey, DHPublicKey,
                             X448PublicKey, X25519PublicKey)) or not pkey:
            # These aren't the keys we're looking for
            return False
            # DSA, RSA
        pd = padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA1()), algorithm=hashes.SHA1(), label=None)
        ciphertext = pkey.public_key().encrypt(Util.RANDOM_DATA, padding=pd)
        refurb = pkey.decrypt(ciphertext, padding=pd)
        return bool(refurb == Util.RANDOM_DATA)

    @staticmethod
    def get_chunks(line_len: int) -> List[Tuple[int, int]]:
        """Returns chunks positions for given line length"""
        # line length is over MAX_LINE_LENGTH already
        chunks = [(0, CHUNK_SIZE)]
        # case for oversize line
        next_offset = CHUNK_STEP_SIZE
        while line_len > next_offset:
            # the target is too long for single "finditer" - it will be scanned by chunks
            if line_len > next_offset + MAX_LINE_LENGTH:
                # the chunk is not the before last
                chunks.append((next_offset, next_offset + CHUNK_SIZE))
                next_offset += CHUNK_STEP_SIZE
            else:
                # the tail of line is between CHUNK_SIZE and MAX_LINE_LENGTH
                chunks.append((next_offset, line_len))
                break
        return chunks

    @staticmethod
    def subtext(text: str, pos: int, hunk_size: int) -> str:
        """cut text symmetrically for given position or use remained quota to be fitted in 2x hunk_size"""
        # cut trailed whitespaces to obtain more informative data
        text = text.rstrip()
        if hunk_size <= pos:
            left_quota = 0
            left_pos = pos - hunk_size
        else:
            left_quota = hunk_size - pos
            left_pos = 0
        # skip leading whitespaces in result string
        for i in range(left_pos, pos):
            if text[i] in string.whitespace:
                left_quota += 1
                left_pos += 1
            else:
                break
        right_remain = len(text) - pos
        if hunk_size <= right_remain:
            right_quota = 0
            right_pos = pos + hunk_size + left_quota
        else:
            right_quota = hunk_size - right_remain
            right_pos = pos + hunk_size + left_quota
        if len(text) < right_pos:
            right_pos = len(text)
        if 0 < left_pos:
            left_pos -= right_quota
            if 0 > left_pos:
                left_pos = 0
        return text[left_pos:right_pos].rstrip()

    @staticmethod
    def get_excel_column_name(column_index: int) -> str:
        """Converts index based column position into Excel style column name"""
        name = ''
        if isinstance(column_index, int):
            while 0 <= column_index:
                column_index, remain = divmod(column_index, 26)
                name = f"{chr(ord('A') + remain)}{name}"
                column_index -= 1
        return name



from abc import abstractmethod
from typing import List, Any, Set, Union

import numpy as np

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.feature import Feature


class WordIn(Feature):
    """Abstract feature returns array with all matched words in a string"""

    def __init__(self, words: List[str]):
        super().__init__()
        self.dimension = len(words)
        self.words = sorted(list(set(words)))
        self.enumerated_words = list(enumerate(self.words))
        if len(self.enumerated_words) != self.dimension:
            raise RuntimeError(f"Check duplicates:{words}")

    @abstractmethod
    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError

    @property
    def zero(self) -> np.ndarray:
        """Returns zero filled array for case of empty input"""
        return np.zeros(shape=[self.dimension], dtype=np.int8)

    def word_in_(self, iterable_data: Union[str, List[str], Set[str]]) -> np.ndarray:
        """Returns array with words included in a string"""
        result: np.ndarray = self.zero
        for i, word in self.enumerated_words:
            if word in iterable_data:
                result[i] = 1
        return np.array([result])



import re
from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValueStringTypeCheck(Filter):
    r"""Check if line_data is in source code file that require quotes for string declaration.

    If it is, then checks if line_data really have string literal declaration.
    Comment rows in source files (start with //, /\*, etc) ignored.
    Multiple bytes scenario allowed [123,23,54,67,78,89] or {0xae, 0x54, 0x55, 0xff}

    True if:

    - line_data have no value
    - line_data have no path
    - line_data is in source code file (.cpp, .py, etc.) and is not comment
      and contain no quotes (so no string literal declared)

    False otherwise
    """

    MULTIBYTE_PATTERN = re.compile(r"((0x)?[0-9a-f]{1,16}[UL]*)(\s*,\s*((0x)?[0-9a-f]{1,16}[UL]*)){3}",
                                   flags=re.IGNORECASE)

    def __init__(self, config: Optional[Config] = None, check_for_literals=True) -> None:
        self.check_for_literals = check_for_literals

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if not self.check_for_literals or line_data.url_part:
            return False

        if ValueStringTypeCheck.MULTIBYTE_PATTERN.search(line_data.value):
            return False

        if line_data.is_source_file_with_quotes() \
                and not line_data.is_comment() \
                and not line_data.is_well_quoted_value \
                and not line_data.is_quoted \
                and not '0' <= line_data.value[0] <= '9' \
                and line_data.separator and '=' in line_data.separator:
            # heterogeneous code e.g. YAML in Python uses colon sign instead equals
            return True

        return False



from abc import ABC, abstractmethod
from typing import List, Any

import numpy as np

from credsweeper.credentials.candidate import Candidate


class Feature(ABC):
    """Base class for features."""

    def __init__(self):
        pass

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        """Call base class for features.

        Args:
            candidates: list of candidates to extract features

        """
        return np.array([self.extract(candidate) for candidate in candidates])

    @abstractmethod
    def extract(self, candidate: Candidate) -> Any:
        """Abstract method of base class"""
        raise NotImplementedError



import contextlib
from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter
from credsweeper.utils.util import Util


class ValueJsonWebKeyCheck(Filter):
    """
    Check that candidate is JWK which starts usually from 'e'
    and have private parts of the key
    https://datatracker.ietf.org/doc/html/rfc7517
    https://datatracker.ietf.org/doc/html/rfc7518
    """

    def __init__(self, config: Optional[Config] = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received key which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        with contextlib.suppress(Exception):
            if data := Util.decode_base64(line_data.value, padding_safe=True, urlsafe_detect=True):
                if b'"kty":' in data and (b'"oct"' in data and b'"k":' in data or
                                          (b'"EC"' in data or b'"RSA"' in data) and b'"d":' in data):
                    return False
        return True



import logging
import lzma
from abc import ABC
from pathlib import Path
from typing import List, Optional, Union

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class LzmaScanner(AbstractScanner, ABC):
    """Implements lzma scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - lzma also xz"""
        if data.startswith((b"\xFD7zXZ\x00", b"\x5D\x00\x00")):
            return True
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data from lzma archive and launches data_scan"""
        try:
            file_path = Path(data_provider.file_path)
            new_path = file_path.as_posix()
            if ".xz" == file_path.suffix:
                new_path = new_path[:-3]
            elif ".lzma" == file_path.suffix:
                new_path = new_path[:-5]
            lzma_content_provider = DataContentProvider(data=lzma.decompress(data_provider.data),
                                                        file_path=new_path,
                                                        file_type=Util.get_extension(new_path),
                                                        info=f"{data_provider.info}|LZMA:{file_path}")
            new_limit = recursive_limit_size - len(lzma_content_provider.data)
            lzma_candidates = self.recursive_scan(lzma_content_provider, depth, new_limit)
            return lzma_candidates
        except Exception as lzma_exc:
            logger.error(f"{data_provider.file_path}:{lzma_exc}")
        return None



import numpy as np

from credsweeper.common.constants import ML_HUNK
from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInPreamble(WordIn):
    """Feature is true if line contains at least one word from predefined list."""

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns true if any words in line before variable or value"""
        if 0 <= candidate.line_data_list[0].variable_start:
            preamble_start = 0 if ML_HUNK >= candidate.line_data_list[0].variable_start \
                else candidate.line_data_list[0].variable_start - ML_HUNK
            preamble = candidate.line_data_list[0].line[preamble_start:candidate.line_data_list[0].
                                                        variable_start].strip()
        else:
            preamble_start = 0 if ML_HUNK >= candidate.line_data_list[0].value_start \
                else candidate.line_data_list[0].value_start - ML_HUNK
            preamble = candidate.line_data_list[0].line[preamble_start:candidate.line_data_list[0].value_start].strip()

        return self.word_in_(preamble.lower()) if preamble else np.array([self.zero])



import numpy as np

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInValue(WordIn):
    """Feature returns true if candidate value contains at least one word from predefined list."""

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns array of matching words for first line"""
        if value := candidate.line_data_list[0].value:
            return self.word_in_(value.lower())
        return np.array([self.zero])



import logging
from dataclasses import dataclass
from functools import cached_property
from typing import List, Tuple, Generator, TypedDict, Optional, Union, Any, Dict

import whatthepatch

from credsweeper.common.constants import DiffRowType
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider

logger = logging.getLogger(__name__)

DiffDict = TypedDict(
    "DiffDict",
    {
        "old": Optional[int],  #
        "new": Optional[int],  #
        "line": Union[str, bytes],  # bytes are possibly since whatthepatch v1.0.4
        "hunk": Any  # not used
    })


@dataclass(frozen=True)
class DiffRowData:
    """Class for keeping data of diff row."""

    line_type: DiffRowType
    line_numb: int
    line: str


class DiffContentProvider(ContentProvider):
    """Provide data from a single `.patch` file.

    Parameters:
        file_path: path to file
        change_type: set added or deleted file data to scan
        diff: list of file row changes, with base elements represented as::

            {
                "old": line number before diff,
                "new": line number after diff,
                "line": line text,
                "hunk": diff hunk number
            }

    """

    def __init__(
            self,  #
            file_path: str,  #
            change_type: DiffRowType,  #
            diff: List[DiffDict]) -> None:
        super().__init__(file_path=file_path, info=f"{file_path}:{change_type.value}")
        self.__change_type = change_type
        self.__diff = diff

    @cached_property
    def data(self) -> bytes:
        """data getter for DiffContentProvider"""
        raise NotImplementedError(__name__)

    @cached_property
    def diff(self) -> List[DiffDict]:
        """diff getter for DiffContentProvider"""
        return self.__diff

    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        self.__diff = []
        if "diff" in self.__dict__:
            delattr(self, "diff")

    @staticmethod
    def parse_lines_data(change_type: DiffRowType, lines_data: List[DiffRowData]) -> Tuple[List[int], List[str]]:
        """Parse diff lines data.

        Return list of line numbers with change type "self.change_type" and list of all lines in file
            in original order(replaced all lines not mentioned in diff file with blank line)

        Args:
            change_type: set added or deleted file data to scan
            lines_data: data of all rows mentioned in diff file

        Return:
            tuple of line numbers with change type "self.change_type" and all file lines
            in original order(replaced all lines not mentioned in diff file with blank line)

        """
        change_numbs = []
        all_lines = []
        for line_data in lines_data:
            if line_data.line_type == change_type:
                change_numbs.append(line_data.line_numb)
                all_lines.append(line_data.line)
        return change_numbs, all_lines

    @staticmethod
    def patch2files_diff(raw_patch: List[str], change_type: DiffRowType) -> Dict[str, List[DiffDict]]:
        """Generate files changes from patch for added or deleted filepaths.

        Args:
            raw_patch: git patch file content
            change_type: change type to select, DiffRowType.ADDED or DiffRowType.DELETED

        Return:
            return dict with ``{file paths: list of file row changes}``, where
            elements of list of file row changes represented as::

                {
                    "old": line number before diff,
                    "new": line number after diff,
                    "line": line text,
                    "hunk": diff hunk number
                }

        """
        if not raw_patch:
            return {}

        added_files, deleted_files = {}, {}
        try:
            for patch in whatthepatch.parse_patch(raw_patch):
                if patch.changes is None:
                    logger.warning(f"Patch '{str(patch.header)}' cannot be scanned")
                    continue
                changes = []
                for change in patch.changes:
                    change_dict = change._asdict()
                    changes.append(change_dict)

                added_files[patch.header.new_path] = changes
                deleted_files[patch.header.old_path] = changes
            if change_type == DiffRowType.ADDED:
                return added_files
            elif change_type == DiffRowType.DELETED:
                return deleted_files
            else:
                logger.error(f"Change type should be one of: '{DiffRowType.ADDED}', '{DiffRowType.DELETED}';"
                             f" but received {change_type}")
        except Exception as exc:
            logger.error(exc)
        return {}

    @staticmethod
    def preprocess_diff_rows(
            added_line_number: Optional[int],  #
            deleted_line_number: Optional[int],  #
            line: str) -> List[DiffRowData]:
        """Auxiliary function to extend diff changes.

        Args:
            added_line_number: number of added line or None
            deleted_line_number: number of deleted line or None
            line: the text line

        Return:
            diff rows data with as list of row change type, line number, row content

        """
        rows_data: List[DiffRowData] = []
        if isinstance(added_line_number, int):
            # indicates line was inserted
            rows_data.append(DiffRowData(DiffRowType.ADDED, added_line_number, line))
        if isinstance(deleted_line_number, int):
            # indicates line was removed
            rows_data.append(DiffRowData(DiffRowType.DELETED, deleted_line_number, line))
        return rows_data

    @staticmethod
    def wrong_change(change: DiffDict) -> bool:
        """Returns True if the change is wrong"""
        for i in ["line", "new", "old"]:
            if i not in change:
                logger.error(f"Skipping wrong change {change}")
                return True
        return False

    @staticmethod
    def preprocess_file_diff(changes: List[DiffDict]) -> List[DiffRowData]:
        """Generate changed file rows from diff data with changed lines (e.g. marked + or - in diff).

        Args:
            changes: git diff by file rows data

        Return:
            diff rows data with as list of row change type, line number, row content

        """
        if not changes:
            return []

        rows_data = []
        # process diff to restore lines and their positions
        for change in changes:
            if DiffContentProvider.wrong_change(change):
                continue
            line = change["line"]
            if isinstance(line, str):
                rows_data.extend(DiffContentProvider.preprocess_diff_rows(change.get("new"), change.get("old"), line))
            elif isinstance(line, (bytes, bytearray)):
                logger.warning("The feature is available with the deep scan option")
            else:
                logger.error(f"Unknown type of line {type(line)}")

        return rows_data

    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Preprocess file diff data to scan.

        Args:
            min_len: minimal line length to scan

        Return:
            list of analysis targets of every row of file diff corresponding to change type "self.change_type"

        """
        lines_data = DiffContentProvider.preprocess_file_diff(self.__diff)
        change_numbs, all_lines = self.parse_lines_data(self.__change_type, lines_data)
        return self.lines_to_targets(min_len, all_lines, change_numbs)



from difflib import SequenceMatcher
from typing import Optional

from credsweeper.common.constants import MIN_VALUE_LENGTH
from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValueSimilarityCheck(Filter):
    """Check if candidate value is over 75% similarity as candidate variable. Like: `secret = "mysecret"` (0.8571)."""

    def __init__(self, config: Optional[Config] = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.variable and line_data.value:
            variable_lower = line_data.variable.lower()
            value_lower = line_data.value.lower()
            if len(value_lower) <= len(variable_lower):
                if value_lower in variable_lower:
                    return True
            elif MIN_VALUE_LENGTH <= len(variable_lower):
                # `api` and `key` may be in the value
                if variable_lower in value_lower:
                    return True
            if 0.75 < SequenceMatcher(None, variable_lower, value_lower).ratio():
                return True
        return False



import logging
from abc import ABC
from typing import List, Optional, Union

from striprtf import striprtf

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class RtfScanner(AbstractScanner, ABC):
    """Implements squash file system scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - Rich Text Format"""
        if data.startswith(b"{\\rtf1") and data.endswith(b"}"):
            return True
        return False

    @staticmethod
    def get_lines(text: str) -> List[str]:
        """Extracts text lines from RTF format"""
        rtf_text = striprtf.rtf_to_text(text)
        lines = Util.split_text(rtf_text)
        return lines

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Scans data as RTF"""
        try:
            string_data_provider = StringContentProvider(lines=RtfScanner.get_lines(data_provider.text),
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|RTF")
            rtf_candidates = self.scanner.scan(string_data_provider)
            return rtf_candidates
        except Exception as rtf_exc:
            logger.error(f"{data_provider.file_path}:{rtf_exc}")
        return None



import contextlib
import datetime
import logging
from abc import abstractmethod, ABC
from typing import List, Optional, Tuple, Any, Generator

from credsweeper.common.constants import RECURSIVE_SCAN_LIMITATION, MIN_DATA_LEN, DEFAULT_ENCODING, UTF_8, \
    MIN_VALUE_LENGTH
from credsweeper.config.config import Config
from credsweeper.credentials.augment_candidates import augment_candidates
from credsweeper.credentials.candidate import Candidate
from credsweeper.file_handler.byte_content_provider import ByteContentProvider
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.descriptor import Descriptor
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider
from credsweeper.file_handler.text_content_provider import TextContentProvider
from credsweeper.scanner.scanner import Scanner

logger = logging.getLogger(__name__)


class AbstractScanner(ABC):
    """Base abstract class for all recursive scanners"""

    @property
    @abstractmethod
    def config(self) -> Config:
        """Abstract property to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @property
    @abstractmethod
    def scanner(self) -> Scanner:
        """Abstract property to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @abstractmethod
    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Abstract method to be defined in DeepScanner"""
        raise NotImplementedError(__name__)

    @staticmethod
    @abstractmethod
    def get_deep_scanners(data: bytes, descriptor: Descriptor, depth: int) -> Tuple[List[Any], List[Any]]:
        """Returns possibly scan methods for the data depends on content and fallback scanners"""
        raise NotImplementedError(__name__)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def recursive_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int = 0,  #
            recursive_limit_size: int = 0) -> List[Candidate]:
        """Recursive function to scan files which might be containers like ZIP archives

            Args:
                data_provider: DataContentProvider object may be a container
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        candidates: List[Candidate] = []
        if 0 > depth:
            # break recursion if maximal depth is reached
            logger.debug("Bottom reached %s recursive_limit_size:%d", data_provider.file_path, recursive_limit_size)
            return candidates
        depth -= 1
        if MIN_DATA_LEN > len(data_provider.data):
            # break recursion for minimal data size
            logger.debug("Too small data: size=%d, depth=%d, limit=%d, path=%s, info=%s", len(data_provider.data),
                         depth, recursive_limit_size, data_provider.file_path, data_provider.info)
            return candidates
        logger.debug("Start data_scan: size=%d, depth=%d, limit=%d, path=%s, info=%s", len(data_provider.data), depth,
                     recursive_limit_size, data_provider.file_path, data_provider.info)

        if FilePathExtractor.is_find_by_ext_file(self.config, data_provider.file_type):
            # Skip scanning file and makes fake candidate due the extension is suspicious
            dummy_candidate = Candidate.get_dummy_candidate(self.config, data_provider.file_path,
                                                            data_provider.file_type, data_provider.info,
                                                            FilePathExtractor.FIND_BY_EXT_RULE)
            candidates.append(dummy_candidate)
        else:
            new_candidates = self.deep_scan_with_fallback(data_provider, depth, recursive_limit_size)
            augment_candidates(candidates, new_candidates)

        return candidates

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def key_value_combination(structure: dict) -> Generator[Tuple[Any, Any], None, None]:
        """Combine items by `key` and `value` from a dictionary for augmentation
        {..., "key": "api_key", "value": "XXXXXXX", ...} -> ("api_key", "XXXXXXX")

        """
        for key_id in ("key", "KEY", "Key"):
            if key_id in structure:
                struct_key = structure.get(key_id)
                break
        else:
            struct_key = None
        if isinstance(struct_key, bytes):
            # sqlite table may produce bytes for `key`
            with contextlib.suppress(UnicodeError):
                struct_key = struct_key.decode(UTF_8)
        # only str type is common used for the augmentation
        if struct_key and isinstance(struct_key, str):
            for value_id in ("value", "VALUE", "Value"):
                if value_id in structure:
                    struct_value = structure.get(value_id)
                    if struct_value and isinstance(struct_value, (str, bytes)):
                        yield struct_key, struct_value
                        # break in successful case
                        break

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def structure_processing(structure: Any) -> Generator[Tuple[Any, Any], None, None]:
        """Yields pair `key, value` from given structure if applicable"""
        if isinstance(structure, dict):
            # transform dictionary to list
            for key, value in structure.items():
                if not value:
                    # skip empty values
                    continue
                if isinstance(value, (list, tuple)):
                    if 1 == len(value):
                        # simplify some structures like YAML when single item in new line is a value
                        yield key, value[0]
                        continue
                # all other data will be precessed in next code
                yield key, value
            yield from AbstractScanner.key_value_combination(structure)
        elif isinstance(structure, (list, tuple)):
            # enumerate the items to fit for return structure
            for key, value in enumerate(structure):
                yield key, value
        else:
            logger.error("Not supported type:%s val:%s", str(type(structure)), repr(structure))

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def structure_scan(
            self,  #
            struct_provider: StructContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> List[Candidate]:
        """Recursive function to scan structured data

            Args:
                struct_provider: DataContentProvider object may be a container
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        candidates: List[Candidate] = []
        logger.debug("Start struct_scan: depth=%d, limit=%d, path=%s, info=%s", depth, recursive_limit_size,
                     struct_provider.file_path, struct_provider.info)

        if 0 > depth:
            # break recursion if maximal depth is reached
            logger.debug("bottom reached %s recursive_limit_size:%d", struct_provider.file_path, recursive_limit_size)
            return candidates

        depth -= 1

        augmented_lines_for_keyword_rules = []
        for key, value in AbstractScanner.structure_processing(struct_provider.struct):
            # a keyword rule may be applicable for `key` (str only) and `value` (str, bytes)
            keyword_match = bool(isinstance(key, str) and self.scanner.keywords_required_substrings_check(key.lower()))

            if isinstance(value, (dict, list, tuple)) and value:
                # recursive scan for not empty structured `value`
                val_struct_provider = StructContentProvider(struct=value,
                                                            file_path=struct_provider.file_path,
                                                            file_type=struct_provider.file_type,
                                                            info=f"{struct_provider.info}|STRUCT:{key}")
                new_candidates = self.structure_scan(val_struct_provider, depth, recursive_limit_size)
                candidates.extend(new_candidates)
            elif isinstance(value, bytes):
                # recursive data scan
                if MIN_DATA_LEN <= len(value):
                    bytes_struct_provider = DataContentProvider(data=value,
                                                                file_path=struct_provider.file_path,
                                                                file_type=struct_provider.file_type,
                                                                info=f"{struct_provider.info}|BYTES:{key}")
                    new_limit = recursive_limit_size - len(value)
                    new_candidates = self.recursive_scan(bytes_struct_provider, depth, new_limit)
                    candidates.extend(new_candidates)
                if keyword_match and MIN_VALUE_LENGTH <= len(value):
                    augmented_lines_for_keyword_rules.append(f"{key} = {repr(value)}")
            elif isinstance(value, str):
                # recursive text scan with transformation into bytes
                stripped_value = value.strip()
                if MIN_DATA_LEN <= len(stripped_value):
                    # recursive scan only for data which may be decoded at least
                    with contextlib.suppress(UnicodeError):
                        data = stripped_value.encode(encoding=DEFAULT_ENCODING, errors='strict')
                        str_struct_provider = DataContentProvider(data=data,
                                                                  file_path=struct_provider.file_path,
                                                                  file_type=struct_provider.file_type,
                                                                  info=f"{struct_provider.info}|STRING:{key}")
                        new_limit = recursive_limit_size - len(str_struct_provider.data)
                        new_candidates = self.recursive_scan(str_struct_provider, depth, new_limit)
                        candidates.extend(new_candidates)
                if keyword_match and MIN_VALUE_LENGTH <= len(stripped_value):
                    augmented_lines_for_keyword_rules.append(f"{key} = {repr(stripped_value)}")
            elif value is None or isinstance(value, (int, float, datetime.date, datetime.datetime)):
                # skip useless types
                pass
            else:
                logger.warning("Not supported type:%s value(%s)", str(type(value)), str(value))

        if augmented_lines_for_keyword_rules:
            str_provider = StringContentProvider(augmented_lines_for_keyword_rules,
                                                 file_path=struct_provider.file_path,
                                                 file_type=struct_provider.file_type,
                                                 info=f"{struct_provider.info}|KEYWORD")
            new_candidates = self.scanner.scan(str_provider)
            augment_candidates(candidates, new_candidates)

        return candidates

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def deep_scan_with_fallback(self, data_provider: DataContentProvider, depth: int,
                                recursive_limit_size: int) -> List[Candidate]:
        """Scans with deep scanners and fallback scanners if possible

            Args:
                data_provider: DataContentProvider with raw data
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack

            Returns: list with candidates

        """
        candidates: List[Candidate] = []
        deep_scanners, fallback_scanners = self.get_deep_scanners(data_provider.data, data_provider.descriptor, depth)
        fallback = True
        for scan_class in deep_scanners:
            new_candidates = scan_class.data_scan(self, data_provider, depth, recursive_limit_size)
            if new_candidates is None:
                # scanner did not recognise the content type
                continue
            augment_candidates(candidates, new_candidates)
            # this scan is successful, so fallback is not necessary
            fallback = False
        if fallback:
            for scan_class in fallback_scanners:
                fallback_candidates = scan_class.data_scan(self, data_provider, depth, recursive_limit_size)
                if fallback_candidates is None:
                    continue
                augment_candidates(candidates, fallback_candidates)
                # use only first successful fallback scanner
                break
        return candidates

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def scan(self,
             content_provider: ContentProvider,
             depth: int,
             recursive_limit_size: Optional[int] = None) -> List[Candidate]:
        """Initial scan method to launch recursive scan. Skips ByteScanner to prevent extra scan

            Args:
                content_provider: ContentProvider that might contain raw data
                depth: maximal level of recursion
                recursive_limit_size: maximal bytes of opened files to prevent recursive zip-bomb attack
        """
        recursive_limit_size = recursive_limit_size if isinstance(recursive_limit_size,
                                                                  int) else RECURSIVE_SCAN_LIMITATION
        candidates: List[Candidate] = []
        data: Optional[bytes] = None
        if isinstance(content_provider, (TextContentProvider, ByteContentProvider)):
            # Feature to scan files which might be containers
            data = content_provider.data
            info = f"FILE:{content_provider.file_path}"
        elif isinstance(content_provider, DiffContentProvider) and content_provider.diff:
            candidates = self.scanner.scan(content_provider)
            # Feature to scan binary diffs
            diff = content_provider.diff[0].get("line")
            # the check for legal fix mypy issue
            if isinstance(diff, bytes):
                data = diff
            info = f"DIFF:{content_provider.file_path}"
        else:
            logger.warning(f"Content provider {type(content_provider)} does not support deep scan")
            info = "NA"

        if data:
            data_provider = DataContentProvider(data=data,
                                                file_path=content_provider.file_path,
                                                file_type=content_provider.file_type,
                                                info=content_provider.info or info)
            new_candidates = self.deep_scan_with_fallback(data_provider, depth, recursive_limit_size - len(data))
            augment_candidates(candidates, new_candidates)
        return candidates



import contextlib
from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter
from credsweeper.utils.util import Util


class ValueBase64KeyCheck(Filter):
    """Check that candidate contains base64 encoded private key"""

    EXTRA_TRANS_TABLE = str.maketrans('', '', "\",'\\")

    def __init__(self, config: Optional[Config] = None) -> None:
        self.config = config

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """

        with contextlib.suppress(Exception):
            # remove backslash escaping sequences
            text = Util.PEM_CLEANING_PATTERN.sub(r'', line_data.value)
            # remove whitespaces
            text = text.translate(Util.WHITESPACE_TRANS_TABLE)
            # clean sequence concatenation case:
            text = text.replace("'+'", '')
            text = text.replace('"+"', '')
            # possibly url based escaping:
            text = text.replace('%2B', '+')
            text = text.replace('%2F', '/')
            text = text.replace('%3D', '=')
            # clean any other chars which should not appear
            text = text.translate(ValueBase64KeyCheck.EXTRA_TRANS_TABLE)
            # only PEM standard encoding supported in regex pattern to cut off ending of the key
            key = Util.decode_base64(text, padding_safe=True, urlsafe_detect=False)
            private_key = Util.load_pk(key, password=None)
            if Util.check_pk(private_key):
                return False
        return True



import io
import logging
from abc import ABC
from typing import List, Optional

from credsweeper.common.constants import DiffRowType
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.patches_provider import PatchesProvider

logger = logging.getLogger(__name__)


class PatchScanner(AbstractScanner, ABC):
    """Implements .patch scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan EML with text representation"""
        try:
            candidates: List[Candidate] = []
            # common limitation
            new_limit_size = recursive_limit_size - len(data_provider.data)
            # ADDED
            path_added = [(data_provider.file_path, io.BytesIO(data_provider.data))]
            added_content_provider = PatchesProvider(path_added, change_type=DiffRowType.ADDED)
            for added_file in added_content_provider.get_scannable_files(self.config):
                added_candidates = self.scan(added_file, depth, new_limit_size)
                candidates.extend(added_candidates)
            # DELETED
            path_deleted = [(data_provider.file_path, io.BytesIO(data_provider.data))]
            deleted_content_provider = PatchesProvider(path_deleted, change_type=DiffRowType.DELETED)
            for deleted_file in deleted_content_provider.get_scannable_files(self.config):
                added_candidates = self.scan(deleted_file, depth, new_limit_size)
                candidates.extend(added_candidates)
            # update the line data for deep scan only
            for i in candidates:
                for line_data in i.line_data_list:
                    line_data.path = f"{data_provider.file_path}/{line_data.path}"
                    line_data.info = f"{data_provider.info}|PATCH:{line_data.info}"
            return candidates
        except Exception as patch_exc:
            logger.error(f"{data_provider.file_path}:{patch_exc}")
        return None



import logging
import re
from abc import ABC
from typing import List, Optional, Union

from credsweeper.common.constants import MAX_LINE_LENGTH
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class XmlScanner(AbstractScanner, ABC):
    """Realises xml scanning"""

    # A well-formed XML must start from < or a whitespace character
    XML_FIRST_BRACKET_PATTERN = re.compile(rb"^\s*<")
    XML_OPENING_TAG_PATTERN = re.compile(rb"<([0-9A-Za-z_]{1,256})")

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """Used to detect xml format from raw bytes"""
        if XmlScanner.XML_FIRST_BRACKET_PATTERN.search(data, 0, MAX_LINE_LENGTH):
            if first_bracket_match := XmlScanner.XML_OPENING_TAG_PATTERN.search(data, 0, MAX_LINE_LENGTH):
                start_pos = first_bracket_match.start()
                closing_tag = b"</" + first_bracket_match.group(1) + b">"
                if start_pos < data.find(closing_tag, start_pos):
                    return True
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to represent data as xml text and scan as text lines"""
        if result := data_provider.represent_as_xml():
            string_data_provider = StringContentProvider(lines=data_provider.lines,
                                                         line_numbers=data_provider.line_numbers,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|XML")
            return self.scanner.scan(string_data_provider)
        return None if result is None else []



import contextlib
from typing import Optional

from credsweeper.common.constants import DEFAULT_PATTERN_LEN, UTF_8
from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter
from credsweeper.utils.util import Util


class ValueBasicAuthCheck(Filter):
    """Check that candidate have a known structure"""

    def __init__(self, config: Optional[Config] = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        value = line_data.value
        with contextlib.suppress(Exception):
            # Basic encoding -> login:password
            decoded = Util.decode_base64(value, padding_safe=True, urlsafe_detect=True)
            delimiter_pos = decoded.find(b':')
            # check whether the delimiter exists and all chars are decoded
            if 0 < delimiter_pos < len(decoded) - DEFAULT_PATTERN_LEN and decoded.decode(UTF_8):
                return False
        return True



import binascii
import contextlib
import logging
import os
import sys
import time
from argparse import ArgumentParser, ArgumentTypeError, Namespace, BooleanOptionalAction
from pathlib import Path
from typing import Any, Union, Dict, Tuple, Sequence

from git import Repo, Commit

from credsweeper import __version__
from credsweeper.app import APP_PATH, CredSweeper
from credsweeper.common.constants import ThresholdPreset, Severity, RuleType, DiffRowType, ML_HUNK
from credsweeper.file_handler.abstract_provider import AbstractProvider
from credsweeper.file_handler.byte_content_provider import ByteContentProvider
from credsweeper.file_handler.files_provider import FilesProvider
from credsweeper.file_handler.patches_provider import PatchesProvider
from credsweeper.logger.logger import Logger
from credsweeper.utils.util import Util

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

logger = logging.getLogger(__name__)


def positive_int(value: Any) -> int:
    """Check if number of parallel processes is not a positive number."""
    int_value = int(value)
    if int_value <= 0:
        logger.error("Number of parallel processes should be a positive number: %s", value)
        raise ArgumentTypeError(f"{value} should be greater than 0")
    return int_value


def threshold_or_float_or_zero(arg: str) -> Union[int, float, ThresholdPreset]:
    """Return ThresholdPreset or a float from the input string

    Args:
        arg: string that either a float or one of allowed values in ThresholdPreset

    Returns:
        int = 0 to disable ML validator, float if arg convertible to float, ThresholdPreset if one of the allowed values

    Raises:
        ArgumentTypeError: if arg cannot be interpreted as float or ThresholdPreset

    """
    allowed_presents = [e.value for e in ThresholdPreset]
    if '0' == arg:
        return 0
    with contextlib.suppress(ValueError):
        return float(arg)  # try convert to float
    if arg in allowed_presents:
        return ThresholdPreset[arg]
    raise ArgumentTypeError(f"value must be a float or one of {allowed_presents}")


def logger_levels(log_level: str) -> str:
    """Logger level correctness verification and transformation

    Args:
        log_level: string with level

    Returns True if log_level UPPERCASE is one of keys
    """
    val = log_level.upper()
    if val in Logger.LEVELS:
        return val
    raise ArgumentTypeError(f"Log level provided: {log_level} -- must be one of: {' | '.join(Logger.LEVELS.keys())}")


def severity_levels(severity_level: str) -> Severity:
    """Severity level correctness verification and transformation

    Args:
        severity_level: string with level

    Returns Severity matched provided string or throws ArgumentTypeError exception
    """

    if severity := Severity.get(severity_level):
        return severity
    raise ArgumentTypeError(
        f"Severity level provided: {severity_level} -- must be one of: {' | '.join([i.value for i in Severity])}")


def check_integrity() -> int:
    """Calculates CRC32 of program files

    Returns CRC32 of files in integer
    """
    crc32 = 0
    for root, _dirs, files in os.walk(APP_PATH):
        for file_name in files:
            if Util.get_extension(file_name) in [".py", ".json", ".txt", ".yaml", ".onnx"]:
                file_path = Path(root) / file_name
                data = Util.read_data(file_path)
                if data:
                    crc32 ^= binascii.crc32(data)
    return crc32


def get_arguments() -> Namespace:
    """All CLI arguments are defined here"""
    parser = ArgumentParser(prog="python -m credsweeper")
    single_banner_argument = 2 == len(sys.argv) and "--banner" == sys.argv[1]
    group = parser.add_mutually_exclusive_group(required=not single_banner_argument)
    group.add_argument("--path", nargs="+", help="file or directory to scan", dest="path", metavar="PATH")
    group.add_argument("--diff_path", nargs="+", help="git diff file to scan", dest="diff_path", metavar="PATH")
    group.add_argument("--export_config",
                       nargs="?",
                       help="exporting default config to file (default: config.json)",
                       const="config.json",
                       dest="export_config",
                       metavar="PATH")
    group.add_argument("--export_log_config",
                       nargs="?",
                       help="exporting default logger config to file (default: log.yaml)",
                       const="log.yaml",
                       dest="export_log_config",
                       metavar="PATH")
    group.add_argument("--git", help="git repo to scan", dest="git", metavar="PATH")
    parser.add_argument("--ref",
                        help="scan git repo from the ref, otherwise - all branches were scanned (slow)",
                        dest="ref",
                        type=str)
    parser.add_argument("--rules",
                        help="path of rule config file (default: credsweeper/rules/config.yaml). "
                        f"severity:{[i.value for i in Severity]} "
                        f"type:{[i.value for i in RuleType]}",
                        default=None,
                        dest="rule_path",
                        metavar="PATH")
    parser.add_argument("--severity",
                        help=f"set minimum level for rules to apply {[i.value for i in Severity]}"
                        f"(default: '{Severity.INFO}', case insensitive)",
                        default=Severity.INFO,
                        dest="severity",
                        type=severity_levels)
    parser.add_argument("--config",
                        help="use custom config (default: built-in)",
                        default=None,
                        dest="config_path",
                        metavar="PATH")
    parser.add_argument("--log_config",
                        help="use custom log config (default: built-in)",
                        default=None,
                        dest="log_config_path",
                        metavar="PATH")
    parser.add_argument("--denylist",
                        help="path to a plain text file with lines or secrets to ignore",
                        default=None,
                        dest="denylist_path",
                        metavar="PATH")
    parser.add_argument("--find-by-ext",
                        help="find files by predefined extension",
                        dest="find_by_ext",
                        action="store_true")
    parser.add_argument("--pedantic",
                        help="process files without extension",
                        action=BooleanOptionalAction,
                        default=False)
    parser.add_argument("--depth",
                        help="additional recursive search in data (experimental)",
                        type=positive_int,
                        dest="depth",
                        default=0,
                        required=False,
                        metavar="POSITIVE_INT")
    parser.add_argument("--no-filters", help="disable filters", dest="no_filters", action="store_false")
    parser.add_argument("--doc", help="document-specific scanning", dest="doc", action="store_true")
    parser.add_argument("--ml_threshold",
                        help="setup threshold for the ml model. "
                        "The lower the threshold - the more credentials will be reported. "
                        f"Allowed values: float between 0 and 1, or any of {[e.value for e in ThresholdPreset]} "
                        "(default: medium)",
                        type=threshold_or_float_or_zero,
                        default=ThresholdPreset.medium,
                        dest="ml_threshold",
                        required=False,
                        metavar="THRESHOLD_OR_FLOAT_OR_ZERO")
    parser.add_argument("--ml_batch_size",
                        "-b",
                        help="batch size for model inference (default: 16)",
                        type=positive_int,
                        dest="ml_batch_size",
                        default=16,
                        required=False,
                        metavar="POSITIVE_INT")
    parser.add_argument("--ml_config",
                        help="use external config for ml model",
                        type=str,
                        default=None,
                        dest="ml_config",
                        required=False,
                        metavar="PATH")
    parser.add_argument("--ml_model",
                        help="use external ml model",
                        type=str,
                        default=None,
                        dest="ml_model",
                        required=False,
                        metavar="PATH")
    parser.add_argument("--ml_providers",
                        help="comma separated list of providers for onnx (CPUExecutionProvider is used by default)",
                        type=str,
                        default=None,
                        dest="ml_providers",
                        required=False,
                        metavar="STR")
    parser.add_argument("--jobs",
                        "-j",
                        help="number of parallel processes to use (default: 1)",
                        type=positive_int,
                        dest="jobs",
                        default=1,
                        metavar="POSITIVE_INT")
    parser.add_argument("--thrifty",
                        help="clear objects after scan to reduce memory consumption",
                        action=BooleanOptionalAction,
                        default=True)
    parser.add_argument("--skip_ignored",
                        help="parse .gitignore files and skip credentials from ignored objects",
                        dest="skip_ignored",
                        action="store_true")
    parser.add_argument("--error",
                        help="produce error code if credentials are found",
                        action=BooleanOptionalAction,
                        default=False)
    parser.add_argument("--save-json",
                        nargs="?",
                        help="save result to json file (default: output.json)",
                        const="output.json",
                        dest="json_filename",
                        metavar="PATH")
    parser.add_argument("--save-xlsx",
                        nargs="?",
                        help="save result to xlsx file (default: output.xlsx)",
                        const="output.xlsx",
                        dest="xlsx_filename",
                        metavar="PATH")
    parser.add_argument("--stdout", help="print results to stdout", action=BooleanOptionalAction, default=True)
    parser.add_argument("--color", help="print results with colorization", action=BooleanOptionalAction, default=False)
    parser.add_argument("--hashed",
                        help="line, variable, value will be hashed in output",
                        action=BooleanOptionalAction,
                        default=False)
    parser.add_argument("--subtext",
                        help=f"line text will be stripped in {2 * ML_HUNK} symbols but value and variable are kept",
                        action=BooleanOptionalAction,
                        default=False)
    parser.add_argument("--sort",
                        help="enable output sorting",
                        dest="sort_output",
                        action=BooleanOptionalAction,
                        default=False)
    parser.add_argument("--log",
                        "-l",
                        help=(f"provide logging level of {list(Logger.LEVELS.keys())}"
                              f" (default: 'warning', case insensitive)"),
                        default="warning",
                        dest="log",
                        metavar="LOG_LEVEL",
                        type=logger_levels)
    parser.add_argument("--size_limit",
                        help="set size limit of files that for scanning (eg. 1GB / 10MiB / 1000)",
                        dest="size_limit",
                        default=None)
    parser.add_argument("--banner",
                        help="show version and crc32 sum of CredSweeper files at start",
                        action="store_const",
                        const=True)
    parser.add_argument("--version",
                        "-V",
                        help="show program's version number and exit",
                        action="version",
                        version=f"CredSweeper {__version__}")
    return parser.parse_args()


def get_credsweeper(args: Namespace) -> CredSweeper:
    """Common function to create the instance"""
    if args.denylist_path is not None:
        denylist = [line for line in Util.read_file(args.denylist_path) if line]
    else:
        denylist = []
    return CredSweeper(rule_path=args.rule_path,
                       config_path=args.config_path,
                       json_filename=args.json_filename,
                       xlsx_filename=args.xlsx_filename,
                       stdout=args.stdout,
                       color=args.color,
                       hashed=args.hashed,
                       subtext=args.subtext,
                       sort_output=args.sort_output,
                       use_filters=args.no_filters,
                       pool_count=args.jobs,
                       ml_batch_size=args.ml_batch_size,
                       ml_threshold=args.ml_threshold,
                       ml_config=args.ml_config,
                       ml_model=args.ml_model,
                       ml_providers=args.ml_providers,
                       find_by_ext=args.find_by_ext,
                       pedantic=args.pedantic,
                       depth=args.depth,
                       doc=args.doc,
                       severity=args.severity,
                       size_limit=args.size_limit,
                       exclude_lines=denylist,
                       exclude_values=denylist,
                       thrifty=args.thrifty,
                       log_level=args.log)


def scan(args: Namespace, content_provider: AbstractProvider) -> int:
    """Scan content_provider data, print results or save them to json_filename is not None

    Args:
        args: arguments of the application
        content_provider: FilesProvider instance to scan data from

    Returns:
        Number of detected credentials

    Warnings:
         DeprecationWarning: Using 'json_filename' and/or 'xlsx_filename' will issue a warning.

    """
    try:
        credsweeper = get_credsweeper(args)
        return credsweeper.run(content_provider=content_provider)
    except Exception as exc:
        logger.critical(exc, exc_info=True)
        logger.exception(exc)
    return -1


def get_commit_providers(commit: Commit, repo: Repo) -> Sequence[ByteContentProvider]:
    """Process a commit and for providers"""
    result = {}
    # use the hardcoded sha1 until sha256 objects are not supported by GitPython
    ancestors = commit.parents or [repo.tree("4b825dc642cb6eb9a060e54bf8d69288fbee4904")]
    for parent in ancestors:
        for diff in parent.diff(commit):
            # only result files
            blob_b = diff.b_blob
            if blob_b and blob_b.path not in result:
                try:
                    result[blob_b.path] = ByteContentProvider(content=blob_b.data_stream.read(),
                                                              file_path=str(blob_b.path),
                                                              info=DiffRowType.ADDED.value)
                except Exception as exc:
                    logger.warning(f"A submodule was not properly initialized or commit was removed: {exc}")
    return list(result.values())


def drill(args: Namespace) -> Tuple[int, int]:
    """Scan repository for branches and commits
    Returns:
        total credentials found
        total scanned commits
    """
    total_credentials = 0
    total_commits = 0
    try:
        # repo init first
        repo = Repo(args.git)
        if args.ref:
            commits_sha1 = set(x.commit.hexsha for x in repo.refs if x.name == args.ref)
            if not commits_sha1:
                commits_sha1 = {args.ref}  # single commit sha1 reference
        else:
            commits_sha1 = set(x.commit.hexsha for x in repo.refs
                               if x.name.startswith('origin/') or x.name.startswith('refs/heads/'))
        logger.info(f"Git repository {args.git} with commits: {commits_sha1}")
        # then - credsweeper
        credsweeper = get_credsweeper(args)
        # use flat iterations to avoid recursive limits
        to_scan = set(commits_sha1)
        # local speedup for already scanned commits - avoid file system interactive
        scanned = set()
        # to avoid double-check
        skipped = set()
        while to_scan:
            commit_sha1 = to_scan.pop()
            if commit_sha1 in scanned:
                # the commit was scanned in this launch
                continue
            commit = repo.commit(commit_sha1)
            if commit.parents:
                # add parents only when they were not skipped or scanned previously
                to_scan.update(x.hexsha for x in commit.parents if x.hexsha not in skipped and x.hexsha not in scanned)
            # check whether the commit has been checked and the report is present
            skip_already_scanned = False
            if args.json_filename:
                json_path = Path(args.json_filename)
                json_path = json_path.with_suffix(f".{commit_sha1}{json_path.suffix}")
                if json_path.exists():
                    skip_already_scanned = True
                else:
                    credsweeper.json_filename = json_path
            if args.xlsx_filename:
                xlsx_path = Path(args.xlsx_filename)
                xlsx_path = xlsx_path.with_suffix(f".{commit_sha1}{xlsx_path.suffix}")
                if xlsx_path.exists():
                    skip_already_scanned = True
                else:
                    credsweeper.xlsx_filename = xlsx_path
            if skip_already_scanned:
                skipped.add(commit_sha1)
                logger.info("Skip already scanned commit: %s %s", commit_sha1, commit.committed_datetime.isoformat())
                continue
            logger.info("Scan commit: %s %s", commit_sha1, commit.committed_datetime.isoformat())
            # prepare all files to scan in the commit with bytes->IO transformation to avoid a multiprocess issue
            if providers := get_commit_providers(commit, repo):
                credsweeper.credential_manager.candidates.clear()
                credsweeper.scan(providers)
                credsweeper.post_processing()
                credsweeper.export_results()
                total_credentials += credsweeper.credential_manager.len_credentials()
            total_commits += 1
            scanned.add(commit_sha1)
    except Exception as exc:
        logger.critical(exc, exc_info=True)
        return -1, total_commits
    return total_credentials, total_commits


def main() -> int:
    """Main function"""
    result = EXIT_FAILURE
    credentials_number = 0
    start_time = time.time()
    args = get_arguments()
    if args.banner:
        print(f"CredSweeper {__version__} crc32:{check_integrity():08x}")
    Logger.init_logging(args.log, args.log_config_path)
    logger.info(f"Init CredSweeper object with arguments: {args} CWD: {os.getcwd()}")
    summary: Dict[str, int] = {}
    if args.path:
        logger.info(f"Run analyzer on path: {args.path}")
        content_provider: AbstractProvider = FilesProvider(args.path, skip_ignored=args.skip_ignored)
        credentials_number = scan(args, content_provider)
        summary["Detected Credentials"] = credentials_number
        if 0 <= credentials_number:
            result = EXIT_SUCCESS
    elif args.diff_path:
        # Analyze added data
        logger.info(f"Run analyzer on added rows from patch files: {args.diff_path}")
        content_provider = PatchesProvider(args.diff_path, change_type=DiffRowType.ADDED)
        add_credentials_number = scan(args, content_provider)
        summary["Added File Credentials"] = add_credentials_number
        # Analyze deleted data
        logger.info(f"Run analyzer on deleted rows from patch files: {args.diff_path}")
        content_provider = PatchesProvider(args.diff_path, change_type=DiffRowType.DELETED)
        del_credentials_number = scan(args, content_provider)
        summary["Deleted File Credentials"] = del_credentials_number
        if 0 <= add_credentials_number and 0 <= del_credentials_number:
            # it means the scan was successful done
            result = EXIT_SUCCESS
            # collect number of all found credential to produce error code when necessary
            credentials_number = add_credentials_number + del_credentials_number
    elif args.git:
        logger.info(f"Run analyzer on GIT: {args.git}")
        credentials_number, commits_number = drill(args)
        summary[f"Detected Credentials in {args.git} for {commits_number} commits "] = credentials_number
        if 0 <= credentials_number:
            result = EXIT_SUCCESS
    elif args.export_config:
        logging.info(f"Exporting default config to file: {args.export_config}")
        config_dict = Util.json_load(APP_PATH / "secret" / "config.json")
        Util.json_dump(config_dict, args.export_config)
        result = EXIT_SUCCESS
    elif args.export_log_config:
        logging.info(f"Exporting default logger config to file: {args.export_log_config}")
        config_dict = Util.yaml_load(APP_PATH / "secret" / "log.yaml")
        Util.yaml_dump(config_dict, args.export_log_config)
        result = EXIT_SUCCESS
    elif args.banner and 2 == len(sys.argv):
        # only extend version invocation
        result = EXIT_SUCCESS
    else:
        logger.error("Not specified 'path' or 'diff_path'")

    if EXIT_SUCCESS == result and len(summary):
        for k, v in summary.items():
            print(f"{k}: {v}")
        end_time = time.time()
        print(f"Time Elapsed: {end_time - start_time}s")

    if args.error and EXIT_SUCCESS == result and 0 < credentials_number:
        # override result when credentials were found with the requirement
        result = EXIT_FAILURE

    return result



import contextlib

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.feature import Feature


class IsSecretNumeric(Feature):
    """Feature is true if candidate value is a numerical value."""

    def extract(self, candidate: Candidate) -> float:
        with contextlib.suppress(ValueError):
            float(candidate.line_data_list[0].value)
            return 1.0
        return -1.0



import logging
import struct
from abc import ABC
from typing import List, Optional, Generator, Tuple, Union

from credsweeper.common.constants import MIN_DATA_LEN, UTF_8
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class DebScanner(AbstractScanner, ABC):
    """Implements deb (ar) scanning"""

    __header_size = 60

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/Deb_(file_format)"""
        if data.startswith(b"!<arch>\n"):
            return True
        return False

    @staticmethod
    def walk_deb(data: bytes) -> Generator[Tuple[int, str, bytes], None, None]:
        """Processes sequence of DEB archive and yields offset, name and data"""
        offset = 8  # b"!<arch>\n"
        data_limit = len(data) - DebScanner.__header_size
        while offset <= data_limit:
            _data = data[offset:offset + DebScanner.__header_size]
            offset += DebScanner.__header_size
            # basic header structure
            _name, _, _size, __ = struct.unpack('16s32s10s2s', _data)
            file_size = int(_size)
            if MIN_DATA_LEN < file_size <= len(data) - offset:
                _data = data[offset:offset + file_size]
                yield offset, _name.decode(encoding=UTF_8).strip().rstrip('/'), _data
            offset += file_size if 0 == 1 & file_size else file_size + 1

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data file from .ar (debian) archive and launches data_scan"""
        try:
            candidates: List[Candidate] = []
            for offset, name, data in DebScanner.walk_deb(data_provider.data):
                deb_content_provider = DataContentProvider(data=data,
                                                           file_path=f"{data_provider.file_path}/{name}",
                                                           file_type=Util.get_extension(name),
                                                           info=f"{data_provider.info}|DEB:0x{offset:x}")
                new_limit = recursive_limit_size - len(data)
                deb_candidates = self.recursive_scan(deb_content_provider, depth, new_limit)
                candidates.extend(deb_candidates)
            return candidates
        except Exception as exc:
            logger.error(exc)
        return None



import io
import logging
from abc import ABC
from typing import List, Optional, Union
from zipfile import ZipFile

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class ZipScanner(AbstractScanner, ABC):
    """Implements zip scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures"""
        if isinstance(data, (bytes, bytearray)) and data.startswith(b"PK") and 4 <= len(data):
            if 0x03 == data[2] and 0x04 == data[3]:
                # normal PK
                return True
            elif 0x05 == data[2] and 0x06 == data[3]:
                # empty archive - no sense to scan in other scanners, so let it be a zip
                return True
            elif 0x07 == data[2] and 0x08 == data[3]:
                # spanned archive - NOT SUPPORTED
                return False
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts files one by one from zip archives and launches data_scan"""
        try:
            candidates = []
            with ZipFile(io.BytesIO(data_provider.data)) as zf:
                for zfl in zf.infolist():
                    # skip directory
                    if zfl.is_dir():
                        continue
                    if FilePathExtractor.check_exclude_file(self.config, zfl.filename):
                        continue
                    if 0 > recursive_limit_size - zfl.file_size:
                        logger.error(f"{zfl.filename}: size {zfl.file_size}"
                                     f" is over limit {recursive_limit_size} depth:{depth}")
                        continue
                    with zf.open(zfl) as f:
                        zip_content_provider = DataContentProvider(data=f.read(),
                                                                   file_path=data_provider.file_path,
                                                                   file_type=Util.get_extension(zfl.filename),
                                                                   info=f"{data_provider.info}|ZIP:{zfl.filename}")
                        # nevertheless use extracted data size
                        new_limit = recursive_limit_size - len(zip_content_provider.data)
                        zip_candidates = self.recursive_scan(zip_content_provider, depth, new_limit)
                        candidates.extend(zip_candidates)
            return candidates
        except Exception as zip_exc:
            # too many exception types might be produced with broken zip
            logger.error(f"{data_provider.file_path}:{zip_exc}")
        return None



from credsweeper.common import static_keyword_checklist
from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.feature import Feature


class MorphemeDense(Feature):
    """Feature calculates morphemes density for a value"""

    def extract(self, candidate: Candidate) -> float:
        density = 0.0
        if value := candidate.line_data_list[0].value.lower():
            morphemes_length = 0
            for morpheme in static_keyword_checklist.morpheme_set:
                morpheme_pos = value.find(morpheme)
                if 0 <= morpheme_pos:
                    morpheme_len = len(morpheme)
                    while 0 <= morpheme_pos:
                        morphemes_length += morpheme_len
                        morpheme_pos += morpheme_len
                        morpheme_pos = value.find(morpheme, morpheme_pos)
            # normalization: minimal morpheme length is 3
            density = morphemes_length / len(value)
            if 1.0 < density:
                # overlap morpheme case
                density = 1.0
        return density



import logging
from abc import ABC
from typing import List, Optional, Union

from credsweeper.common.constants import MAX_LINE_LENGTH
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class HtmlScanner(AbstractScanner, ABC):
    """Implements html scanning if possible"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """Used to detect html format. Suppose, invocation of is_xml() was True before."""
        for opening_tag, closing_tag in [(b"<html", b"</html>"), (b"<body", b"</body>"), (b"<table", b"</table>"),
                                         (b"<p>", b"</p>"), (b"<span>", b"</span>"), (b"<div>", b"</div>"),
                                         (b"<li>", b"</li>"), (b"<ol>", b"</ol>"), (b"<ul>", b"</ul>"),
                                         (b"<th>", b"</th>"), (b"<tr>", b"</tr>"), (b"<td>", b"</td>")]:
            opening_pos = data.find(opening_tag, 0, MAX_LINE_LENGTH)
            if 0 <= opening_pos < data.find(closing_tag, opening_pos):
                # opening and closing tags were found - suppose it is an HTML
                return True
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to represent data as html text and scan as text lines"""
        if result := data_provider.represent_as_html(depth, recursive_limit_size,
                                                     self.scanner.keywords_required_substrings_check):
            string_data_provider = StringContentProvider(lines=data_provider.lines,
                                                         line_numbers=data_provider.line_numbers,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|HTML")
            return self.scanner.scan(string_data_provider)
        return None if result is None else []



from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValueSearchCheck(Filter):
    """Check whether a candidate value contains a pattern - useful for multi rules"""

    def __init__(self, config: Optional[Config] = None, pattern: Optional[str] = None) -> None:
        self.pattern = pattern

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if self.pattern and line_data.value:
            if len(self.pattern) < len(line_data.value):
                if self.pattern in line_data.value:
                    return True
            else:
                if line_data.value in self.pattern:
                    return True
        return False



import copy
import re
from typing import List

from credsweeper.common.constants import RuleType, MAX_LINE_LENGTH
from credsweeper.config.config import Config
from credsweeper.credentials.candidate import Candidate
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import ValueSearchCheck
from credsweeper.filters.filter import Filter
from credsweeper.rules.rule import Rule
from credsweeper.scanner.scan_type.scan_type import ScanType


class MultiPattern(ScanType):
    """Check if line is a part of a multi-line credential and second part is present within MAX_SEARCH_MARGIN lines.

    Parameters:
        MAX_SEARCH_MARGIN: Int constant. Number of lines around current to perform search for the second part

    """

    MAX_SEARCH_MARGIN = 10

    @classmethod
    def run(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Check if multiline credential present if the file within MAX_SEARCH_MARGIN range from current line_num.

        Args:
            config: user configs
            rule: Rule object to check current line. Should be a multi-pattern rule
            target: Analysis target

        Return:
            List of Candidates if pattern defined in a rule is present in a line
            and second part of multi-pattern rule is present within MAX_SEARCH_MARGIN from the line.
            Empty list (False) - otherwise.

        """
        if RuleType.MULTI != rule.rule_type:
            raise ValueError(f"Rule `{rule}` provided to `{cls.__name__}`.run "
                             f"should have pattern_type equal to `{RuleType.MULTI.value}`")

        candidates = cls._get_candidates(config, rule, target)

        for candidate in candidates:
            # use additional filter to skip the value in first line_data and continues scan
            filters = copy.deepcopy(rule.filters)
            filters.append(ValueSearchCheck(config, candidate.line_data_list[0].value))
            for line_pos in cls.get_line_positions(candidate.line_data_list[0].line_pos, target):
                if cls._scan(config, candidate, line_pos, target, rule.patterns[1], filters):
                    break

        # return candidates with multi line_data_list only
        return [x for x in candidates if 1 < len(x.line_data_list)]

    @classmethod
    def get_line_positions(cls, line_pos: int, target: AnalysisTarget) -> List[int]:
        """Returns list of line positions to be scanned for second part of multi-pattern rule in a priority order."""
        if 0 <= line_pos < target.lines_len:
            # the same line is first
            priority_positions = [(0, line_pos)]
        else:
            return []
        # margin order is constant at start
        priority_forward = priority_backward = cls.MAX_SEARCH_MARGIN
        # backward lines are second priority
        priority_backward += cls.MAX_SEARCH_MARGIN
        line_pos_margin = 1
        while line_pos_margin <= cls.MAX_SEARCH_MARGIN:
            # forward
            line_pos_forward = line_pos + line_pos_margin
            if 0 <= line_pos_forward < target.lines_len:
                if forward_curled_diff := target.lines[line_pos_forward].count('}', 0, MAX_LINE_LENGTH):
                    forward_curled_diff -= target.lines[line_pos_forward].count('{', 0, MAX_LINE_LENGTH)
                if 0 < forward_curled_diff:
                    priority_forward += cls.MAX_SEARCH_MARGIN * (1 + forward_curled_diff)
                else:
                    priority_forward += cls.MAX_SEARCH_MARGIN
                priority_positions.append((priority_forward, line_pos_forward))
            # backward
            line_pos_backward = line_pos - line_pos_margin
            if 0 <= line_pos_backward < target.lines_len:
                if backward_curled_diff := target.lines[line_pos_backward].count('{', 0, MAX_LINE_LENGTH):
                    backward_curled_diff -= target.lines[line_pos_backward].count('}', 0, MAX_LINE_LENGTH)
                if 0 < backward_curled_diff:
                    priority_backward += cls.MAX_SEARCH_MARGIN * (1 + backward_curled_diff)
                else:
                    priority_backward += cls.MAX_SEARCH_MARGIN
                priority_positions.append((priority_backward, line_pos_backward))
            # increment the margin for next index
            line_pos_margin += 1
        # first item is priority, second - line_pos
        priority_positions.sort()
        return [x for _, x in priority_positions]

    @classmethod
    def _scan(cls, config: Config, candidate: Candidate, candi_line_pos: int, target: AnalysisTarget,
              pattern: re.Pattern, filters: List[Filter]) -> bool:
        """Search for second pattern in multi-pattern rule.

        Automatically update candidate with detected line if any.

        Args:
            config: dict, scanner configuration
            candidate: Current credential candidate detected in the line
            candi_line_pos: line position of lines around candidate to perform search
            target: Analysis target
            pattern: second pattern in a rule
            filters: filters to be applied on candidate

        Return:
            Boolean. True if second part detected. False otherwise

        """
        new_target = AnalysisTarget(candi_line_pos, target.lines, target.line_nums, target.descriptor)

        line_data_list = cls.get_line_data_list(config=config, target=new_target, pattern=pattern, filters=filters)

        if not line_data_list:
            return False
        else:
            candidate.line_data_list.extend(line_data_list)
            return True



import logging
import re
from typing import List, Any, Tuple, Union, Dict

from credsweeper.common.constants import MIN_DATA_LEN
from credsweeper.config.config import Config
from credsweeper.deep_scanner.byte_scanner import ByteScanner
from credsweeper.deep_scanner.bzip2_scanner import Bzip2Scanner
from credsweeper.deep_scanner.csv_scanner import CsvScanner
from credsweeper.deep_scanner.deb_scanner import DebScanner
from credsweeper.deep_scanner.docx_scanner import DocxScanner
from credsweeper.deep_scanner.eml_scanner import EmlScanner
from credsweeper.deep_scanner.encoder_scanner import EncoderScanner
from credsweeper.deep_scanner.gzip_scanner import GzipScanner
from credsweeper.deep_scanner.html_scanner import HtmlScanner
from credsweeper.deep_scanner.jclass_scanner import JclassScanner
from credsweeper.deep_scanner.jks_scanner import JksScanner
from credsweeper.deep_scanner.lang_scanner import LangScanner
from credsweeper.deep_scanner.lzma_scanner import LzmaScanner
from credsweeper.deep_scanner.mxfile_scanner import MxfileScanner
from credsweeper.deep_scanner.patch_scanner import PatchScanner
from credsweeper.deep_scanner.pdf_scanner import PdfScanner
from credsweeper.deep_scanner.pkcs_scanner import PkcsScanner
from credsweeper.deep_scanner.png_scanner import PngScanner
from credsweeper.deep_scanner.pptx_scanner import PptxScanner
from credsweeper.deep_scanner.rpm_scanner import RpmScanner
from credsweeper.deep_scanner.rtf_scanner import RtfScanner
from credsweeper.deep_scanner.sqlite3_scanner import Sqlite3Scanner
from credsweeper.deep_scanner.strings_scanner import StringsScanner
from credsweeper.deep_scanner.tar_scanner import TarScanner
from credsweeper.deep_scanner.tmx_scanner import TmxScanner
from credsweeper.deep_scanner.xlsx_scanner import XlsxScanner
from credsweeper.deep_scanner.xml_scanner import XmlScanner
from credsweeper.deep_scanner.zip_scanner import ZipScanner
from credsweeper.deep_scanner.zlib_scanner import ZlibScanner
from credsweeper.file_handler.descriptor import Descriptor
from credsweeper.scanner.scanner import Scanner
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class DeepScanner(
    ByteScanner,  #
    Bzip2Scanner,  #
    DocxScanner,  #
    CsvScanner,  #
    EncoderScanner,  #
    GzipScanner,  #
    HtmlScanner,  #
    JclassScanner,  #
    JksScanner,  #
    LangScanner,  #
    LzmaScanner,  #
    MxfileScanner,  #
    EmlScanner,  #
    PatchScanner,  #
    PdfScanner,  #
    PkcsScanner,  #
    PngScanner,  #
    PptxScanner,  #
    RtfScanner,  #
    RpmScanner,  #
    Sqlite3Scanner,  #
    StringsScanner,  #
    TarScanner,  #
    DebScanner,  #
    XmlScanner,  #
    XlsxScanner,  #
    ZipScanner,  #
    ZlibScanner,  #
):  # yapf: disable
    """Advanced scanner with recursive exploring of data"""

    def __init__(self, config: Config, scanner: Scanner) -> None:
        """Initialize Advanced credential scanner.

        Args:
            scanner: CredSweeper scanner object
            config: dictionary variable, stores analyzer features
        """
        self.__config = config
        self.__scanner = scanner

    @property
    def config(self) -> Config:
        return self.__config

    @property
    def scanner(self) -> Scanner:
        return self.__scanner

    # manually crafted dict to detect a media format with first byte, prefix and optionally pattern
    MEDIA_PATTERNS: Dict[int, List[Tuple[bytes, re.Pattern]]] = {
        0x00: [
            (b"\x00\x00\x00\x0C\x6A\x50\x20\x20\x0D\x0A\x87\x0A", None),  # JPEG2000
            (b"\x00\x00\x01\x00", None),  # ICO
        ],
        0x1A: [
            (b"\x1A\x45\xDF\xA3", None),  # Matroska
        ],
        0x89: [
            (b"\x89PNG\x0D\x0A\x1A\x0A", None),  # PNG - can store text chunks inside
        ],
        0xFF: [
            (b"\xFF", re.compile(b"\xFF(\xD8\xFF[\xDB\xEE\xE1\xE0\x51]|[\xFB\xF3\xF2])")),  # JPEG or MPEG-1 Layer 3
        ],
        ord('B'): [
            (b"BM", re.compile(b"BM.{2}\x00{4}")),  # BMP
        ],
        ord('G'): [
            (b"GIF8", re.compile(b"GIF8[79]a.{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # GIF
        ],
        ord('I'): [
            (b"II", re.compile(b"II[+*]\x00.{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # TIFF little endian
        ],
        ord('M'): [
            (b"MM", re.compile(b"MM\x00[+*].{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # TIFF big endian
        ],
        ord('O'): [
            (b"OggS", re.compile(b"OggS.{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # OGG
        ],
        ord('R'): [
            (b"RIF", re.compile(b"RIF[FX].{4}[ 0-9A-Za-z]{4}.{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # RIFF va
        ],
        ord('X'): [
            (b"XFIR", re.compile(b"XFIR.{4}[ 0-9A-Za-z]{4}.{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # Macromedia
        ],
        ord('f'): [
            (b"ftyp", re.compile(b"ftyp(isom|MSNV).{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # mp4
        ],
        ord('w'): [
            (b"wOF", re.compile(b"wOF[2F].{0,4096}[\x00-\x08\x0C\x0E\x1F\x80-\xFF]")),  # WOFF 1.0, 2.0
        ],
    }

    @staticmethod
    def is_media(data: Union[bytes, bytearray]) -> bool:
        """Returns True if well-known media format found"""
        if patterns := DeepScanner.MEDIA_PATTERNS.get(data[0]):
            for prefix, pattern in patterns:
                # use prefix for speed-up total search
                if prefix and data.startswith(prefix) and (pattern is None or pattern.match(data)):
                    return True
        return False

    @staticmethod
    def get_deep_scanners(data: bytes, descriptor: Descriptor, depth: int) -> Tuple[List[Any], List[Any]]:
        """Returns possibly scan methods for the data depends on content and fallback scanners"""
        deep_scanners: List[Any] = []
        fallback_scanners: List[Any] = []
        if not data or not isinstance(data, (bytes, bytearray)) or len(data) < MIN_DATA_LEN:
            # Guard clause: reject empty or invalid input data early
            pass
        elif ZipScanner.match(data):
            if 0 < depth:
                deep_scanners.append(ZipScanner)
            # probably, there might be a docx, xlsx and so on.
            # It might be scanned with text representation in third-party libraries.
            if descriptor.extension in (".xlsx", ".ods"):
                deep_scanners.append(XlsxScanner)
            else:
                fallback_scanners.append(XlsxScanner)
            if ".docx" == descriptor.extension:
                deep_scanners.append(DocxScanner)
            else:
                fallback_scanners.append(DocxScanner)
            if ".pptx" == descriptor.extension:
                deep_scanners.append(PptxScanner)
            else:
                fallback_scanners.append(PptxScanner)
        elif XlsxScanner.match(data):
            if ".xls" == descriptor.extension:
                deep_scanners.append(XlsxScanner)
            else:
                fallback_scanners.append(XlsxScanner)
        elif Bzip2Scanner.match(data):
            if 0 < depth:
                deep_scanners.append(Bzip2Scanner)
        elif LzmaScanner.match(data):
            if 0 < depth:
                deep_scanners.append(LzmaScanner)
        elif TarScanner.match(data):
            if 0 < depth:
                deep_scanners.append(TarScanner)
        elif DebScanner.match(data):
            if 0 < depth:
                deep_scanners.append(DebScanner)
        elif GzipScanner.match(data):
            if 0 < depth:
                deep_scanners.append(GzipScanner)
        elif PdfScanner.match(data):
            deep_scanners.append(PdfScanner)
        elif PngScanner.match(data):
            deep_scanners.append(PngScanner)
        elif RpmScanner.match(data):
            if 0 < depth:
                deep_scanners.append(RpmScanner)
        elif JclassScanner.match(data):
            deep_scanners.append(JclassScanner)
        elif JksScanner.match(data):
            deep_scanners.append(JksScanner)
        elif Sqlite3Scanner.match(data):
            if 0 < depth:
                deep_scanners.append(Sqlite3Scanner)
        elif PkcsScanner.match(data):
            deep_scanners.append(PkcsScanner)
        elif RtfScanner.match(data):
            deep_scanners.append(RtfScanner)
            fallback_scanners.append(ByteScanner)
        elif XmlScanner.match(data):
            if HtmlScanner.match(data):
                deep_scanners.append(HtmlScanner)
                deep_scanners.append(XmlScanner)
                fallback_scanners.append(ByteScanner)
            elif MxfileScanner.match(data):
                deep_scanners.append(MxfileScanner)
                deep_scanners.append(XmlScanner)
                fallback_scanners.append(ByteScanner)
            elif TmxScanner.match(data):
                deep_scanners.append(TmxScanner)
                fallback_scanners.append(XmlScanner)
                fallback_scanners.append(ByteScanner)
            else:
                deep_scanners.append(XmlScanner)
                fallback_scanners.append(ByteScanner)
        elif EmlScanner.match(data):
            if descriptor.extension in (".eml", ".mht"):
                deep_scanners.append(EmlScanner)
            else:
                if 0 < depth:
                    # a formal patch looks like an eml
                    deep_scanners.append(PatchScanner)
                fallback_scanners.append(EmlScanner)
            fallback_scanners.append(ByteScanner)
        elif DeepScanner.is_media(data):
            # only StringsScanner may be applied for the formats effective
            if 0 < depth:
                fallback_scanners.append(StringsScanner)
        elif not Util.is_binary(data):
            # keep ByteScanner first to apply real value position if possible
            deep_scanners.append(ByteScanner)
            if 0 < depth:
                deep_scanners.append(PatchScanner)
                deep_scanners.append(LangScanner)
                if CsvScanner.match(data):
                    deep_scanners.append(CsvScanner)
                if EncoderScanner.match(data):
                    deep_scanners.append(EncoderScanner)
                if ZlibScanner.match(data):
                    deep_scanners.append(ZlibScanner)
        else:
            if 0 < depth:
                if ZlibScanner.match(data):
                    deep_scanners.append(ZlibScanner)
                    fallback_scanners.append(StringsScanner)
                else:
                    deep_scanners.append(StringsScanner)
            else:
                logger.warning("Cannot apply a deep scanner for type %s prefix %s %d", descriptor, repr(data[:32]),
                               len(data))
        return deep_scanners, fallback_scanners



import math
from functools import cache

from credsweeper.filters.value_entropy_base_check import ValueEntropyBaseCheck


class ValueEntropyBase36Check(ValueEntropyBaseCheck):
    """Base36 entropy check"""

    @staticmethod
    @cache
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal entropy for size of random data. Precalculated data is applied for speedup"""
        if 15 == x:
            # workaround for Dropbox App secret
            y = 3.374
        elif 10 <= x < 26:
            y = 0.731566857 * math.log2(x) + 0.474132
        elif 26 <= x:
            y = 3.9
        else:
            y = 0
        return y



from typing import Optional

from credsweeper.common import static_keyword_checklist
from credsweeper.common.constants import MAX_LINE_LENGTH
from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValueMorphemesCheck(Filter):
    """Check value for a threshold of morphemes count"""

    THRESHOLDS_X3 = int(MAX_LINE_LENGTH).bit_length()
    # one morpheme is very likely to be random generated even for 3 symbols
    MAX_MORPHEMES_LIMIT = max(1, THRESHOLDS_X3 - 4)

    def __init__(self, config: Optional[Config] = None, threshold: Optional[int] = None) -> None:
        # threshold - minimum morphemes number in a value
        if threshold is None:
            # use dynamic thresholds
            self.thresholds = [max(1, x - 4) for x in range(ValueMorphemesCheck.THRESHOLDS_X3)]
        elif isinstance(threshold, int) and 0 <= threshold:
            # constant thresholds for any pattern
            self.thresholds = [threshold] * ValueMorphemesCheck.THRESHOLDS_X3
        else:
            raise ValueError(f"Wrong type of pattern length {type(threshold)} = {repr(threshold)}")

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        threshold_id = len(line_data.value).bit_length()
        # use the last (max) threshold in very huge value
        threshold = self.thresholds[threshold_id] if len(self.thresholds) > threshold_id else self.thresholds[-1]
        return static_keyword_checklist.check_morphemes(line_data.value.lower(), threshold)



import logging
import zlib
from abc import ABC
from typing import List, Optional

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider

logger = logging.getLogger(__name__)


class ZlibScanner(AbstractScanner, ABC):
    """Implements zlib data inflate and scan"""

    @staticmethod
    def match(data: bytes) -> bool:
        """Returns True if data looks like deflated data with zlib"""
        if 6 < len(data):
            cmf = data[0]
            flg = data[1]
            if 8 == (0xF & cmf) and 7 >= (cmf >> 4) and 0 == ((cmf << 8) | flg) % 31 and 0 == (0x20 & flg):
                if 0x3 != (data[2] >> 1):
                    # the last check of impossible bits
                    return True
        return False

    @staticmethod
    def decompress(limit: int, data: bytes) -> bytes:
        """Returns decompressed data by chunks with a limit or exception in unusual cases"""
        zlib_obj = zlib.decompressobj()
        result = zlib_obj.decompress(data, max_length=limit)
        if zlib_obj.unconsumed_tail:
            raise ValueError(f"Limit exceeds for {len(zlib_obj.unconsumed_tail)}")
        if not zlib_obj.eof:
            raise ValueError("Truncated zlib stream")
        if zlib_obj.unused_data:
            raise ValueError(f"Unused data {len(zlib_obj.unused_data)}")
        return result

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Inflate data from zlib compressed and launches data_scan"""
        try:
            decompressed = ZlibScanner.decompress(recursive_limit_size, data_provider.data)
            zlib_content_provider = DataContentProvider(data=decompressed,
                                                        file_path=data_provider.file_path,
                                                        file_type=data_provider.file_type,
                                                        info=f"{data_provider.info}|ZLIB")
            new_limit = recursive_limit_size - len(decompressed)
            zlib_candidates = self.recursive_scan(zlib_content_provider, depth, new_limit)
            return zlib_candidates
        except Exception as zlib_exc:
            logger.error(f"{data_provider.file_path}:{zlib_exc}")
        return None



from credsweeper.common.constants import CHUNK_SIZE
from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn
from credsweeper.utils.util import Util


class HasHtmlTag(WordIn):
    """Feature is true if line has HTML tags (HTML file)."""

    HTML_WORDS = [
        '< img', '<img', '< script', '<script', '< p', '<p', '< link', '<link', '< meta', '<meta', '< a', '<a'
    ]

    def __init__(self) -> None:
        super().__init__(HasHtmlTag.HTML_WORDS)

    def extract(self, candidate: Candidate) -> float:
        subtext = Util.subtext(candidate.line_data_list[0].line, candidate.line_data_list[0].value_start, CHUNK_SIZE)
        candidate_line_data_list_0_line_lower = subtext.lower()
        if '<' not in candidate_line_data_list_0_line_lower:
            # early check
            return -1.0
        for i in self.words:
            if i in candidate_line_data_list_0_line_lower:
                return 1.0
        if "/>" in candidate_line_data_list_0_line_lower or "</" in candidate_line_data_list_0_line_lower:
            # possible closed tag
            return 1.0
        return -1.0



import base64
import logging
from abc import ABC
from typing import List, Optional, Union

from credsweeper.common.constants import Severity, Confidence
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class PkcsScanner(AbstractScanner, ABC):
    """Implements pkcs12 scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> int:
        """Matched ASN1 structure"""
        return bool(Util.get_asn1_size(data))

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan PKCS12 to open with standard password"""
        for pw_probe in self.config.bruteforce_list:
            try:
                password = pw_probe.encode() if pw_probe else None
                if pkey := Util.load_pk(data_provider.data, password):
                    if not Util.check_pk(pkey):
                        logger.debug("False alarm %s", data_provider.info)
                        return []
                    candidate = Candidate.get_dummy_candidate(
                        self.config,  #
                        data_provider.file_path,  #
                        data_provider.file_type,  #
                        info=f"{data_provider.info}|PKCS_PASSWORD:{repr(password)}",  #
                        rule_name=f"PKCS with password {repr(pw_probe)}" if pw_probe else "PKCS without password")
                    candidate.line_data_list[0].line = base64.b64encode(data_provider.data).decode()
                    candidate.line_data_list[0].value = repr(password)
                    # high severity is assigned to private key rules
                    candidate.severity = Severity.HIGH
                    candidate.confidence = Confidence.STRONG
                    return [candidate]
            except Exception as pkcs_exc:
                logger.debug(f"{data_provider.file_path}:{pw_probe}:{pkcs_exc}")
        return None



import numpy as np

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInVariable(WordIn):
    """Feature returns array of words matching in variable"""

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns array of matching words for first line"""
        if variable := candidate.line_data_list[0].variable:
            return self.word_in_(variable.lower())
        return np.array([self.zero])



import csv
import io
import logging
import re
from abc import ABC
from typing import List, Optional, Dict, Any

from credsweeper.common.constants import MAX_LINE_LENGTH
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider

logger = logging.getLogger(__name__)


class CsvScanner(AbstractScanner, ABC):
    """Implements CSV scanning"""

    sniffer = csv.Sniffer()
    # do not use space as separator to avoid hallucinations
    DELIMITERS = ",;\t|\x1F"

    CSV_PATTERN = re.compile(b"[^\r\n]{1,8000}[,;\t|\x1F][^\r\n]{1,8000}")

    @staticmethod
    def match(data: bytes) -> bool:
        """Check if data MAY be in CSV format"""
        end_pos = data.find(b'\n', 0, MAX_LINE_LENGTH)
        if 0 > end_pos:
            # classic Mac OS format
            end_pos = data.find(b'\r', 0, MAX_LINE_LENGTH)
        if 0 <= end_pos:
            if CsvScanner.CSV_PATTERN.match(data, pos=0, endpos=end_pos):
                return True
        return False

    @classmethod
    def get_structure(cls, text: str) -> List[Dict[str, Any]]:
        """Reads a text as CSV standard with guessed dialect"""
        # windows style \r\n
        first_line_end = text.find('\r', 0, MAX_LINE_LENGTH)
        line_terminator = "\r\n"
        if 0 > first_line_end:
            # unix style \n
            first_line_end = text.find('\n', 0, MAX_LINE_LENGTH)
            line_terminator = "\n"
            if 0 > first_line_end:
                raise ValueError(f"No suitable line end found in {MAX_LINE_LENGTH} symbols")

        first_line = text[:first_line_end]
        dialect = cls.sniffer.sniff(first_line, delimiters=cls.DELIMITERS)
        rows = []
        reader = csv.DictReader(io.StringIO(text),
                                delimiter=dialect.delimiter,
                                lineterminator=line_terminator,
                                strict=True)
        # check the constant columns number for all rows
        fields_number = sum(1 for x in reader.fieldnames if x is not None)
        for row in reader:
            if not isinstance(row, dict):
                raise ValueError(f"ERROR: wrong row '{row}'")
            if len(row) != fields_number or any(x is None for x in row.values()):
                # None means no separator used
                raise ValueError(f"Different columns number in row '{row}' - mismatch {fields_number}")
            rows.append(row)
        return rows

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan each row as structure with column name in key"""
        try:
            if rows := self.get_structure(data_provider.text):
                struct_content_provider = StructContentProvider(struct=rows,
                                                                file_path=data_provider.file_path,
                                                                file_type=data_provider.file_type,
                                                                info=f"{data_provider.info}|CSV")
                new_limit = recursive_limit_size - sum(len(x) for x in rows)
                struct_candidates = self.structure_scan(struct_content_provider, depth, new_limit)
                return struct_candidates
        except Exception as csv_exc:
            logger.debug(f"{data_provider.file_path}:{csv_exc}")
        return None



from abc import abstractmethod
from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter
from credsweeper.utils.util import Util


class ValueEntropyBaseCheck(Filter):
    """Check that candidate value has minimal Shanon Entropy for appropriated base"""

    def __init__(self, config: Optional[Config] = None) -> None:
        pass

    @staticmethod
    @abstractmethod
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal entropy for size of data"""
        raise NotImplementedError()

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        entropy = Util.get_shannon_entropy(line_data.value)
        min_entropy = self.get_min_data_entropy(len(line_data.value))
        if min_entropy > entropy or 0 == min_entropy:
            return True
        return False



import numpy as np

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInTransition(WordIn):
    """Feature is true if line contains at least one word from predefined list."""

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns true if any words between variable and value"""
        if 0 <= candidate.line_data_list[0].variable_end < candidate.line_data_list[0].value_start:
            transition = candidate.line_data_list[0].line[candidate.line_data_list[0].variable_end:candidate.
                                                          line_data_list[0].value_start].strip()
        else:
            transition = ''

        return self.word_in_(transition.lower()) if transition else np.array([self.zero])



import math
from functools import cache

from credsweeper.filters.value_entropy_base_check import ValueEntropyBaseCheck


class ValueEntropyBase64Check(ValueEntropyBaseCheck):
    """Base64 entropy check"""

    @staticmethod
    @cache
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal average entropy for size of random data. Precalculated round data is applied for speedup"""
        if 12 <= x < 18:
            y = 0.915 * math.log2(x) - 0.047
        elif 18 <= x < 35:
            y = 0.767 * math.log2(x) + 0.5677
        elif 35 <= x < 65:
            y = 0.944 * math.log2(x) - 0.009 * x - 0.04
        elif 65 <= x < 256:
            y = 0.621 * math.log2(x) - 0.003 * x + 1.54
        elif 256 <= x:
            y = 6 - 64 / x
        else:
            y = 0
        return y



import os.path
from pathlib import Path
from typing import List, Any

import numpy as np

from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInPath(WordIn):
    """Categorical feature that corresponds to words in path (POSIX, lowercase)"""

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        # actually there must be one path because the candidates are grouped before
        if file_path := candidates[0].line_data_list[0].path:
            path = Path(file_path)
            # apply ./ for normalised path to detect "/src" for relative path
            posix_lower_path = path.as_posix().lower() if path.is_absolute() else f"./{path.as_posix().lower()}"
            # prevent extra confusion from the same word in extension
            path_without_extension, _ = os.path.splitext(posix_lower_path)
            return self.word_in_(path_without_extension)
        return np.array([self.zero])

    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError



from credsweeper.common.constants import GroupType
from credsweeper.config.config import Config
from credsweeper.filters import ValueMorphemesCheck, ValuePatternCheck, ValueNumberCheck, ValueEntropyBase36Check, \
    ValueTokenBase36Check
from credsweeper.filters.group.group import Group


class WeirdBase36Token(Group):
    """Structured Token"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueMorphemesCheck(threshold=1),
            ValuePatternCheck(),
            ValueNumberCheck(),
            ValueTokenBase36Check(),
            ValueEntropyBase36Check(),
        ]



from credsweeper.common.constants import GroupType
from credsweeper.config.config import Config
from credsweeper.filters import (ValueAllowlistCheck, ValueArrayDictionaryCheck, ValueBlocklistCheck,
                                 ValueCamelCaseCheck, ValueLengthCheck, ValueFilePathCheck, ValueLastWordCheck,
                                 ValueMethodCheck, ValueNotAllowedPatternCheck, ValuePatternCheck, ValueStringTypeCheck,
                                 ValueTokenCheck)
from credsweeper.filters.group.group import Group


class UrlCredentialsGroup(Group):
    """UrlCredentialsGroup"""

    def __init__(self, config: Config) -> None:
        """URL credentials group class.

        Similar to PasswordKeyword, but exclude all checks dependent on the variable name, as URL credentials have no
        explicitly defined variable
        """
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueAllowlistCheck(),
            ValueArrayDictionaryCheck(),
            ValueBlocklistCheck(),
            ValueCamelCaseCheck(),
            ValueFilePathCheck(),
            ValueLastWordCheck(),
            ValueMethodCheck(),
            ValueStringTypeCheck(check_for_literals=config.check_for_literals),
            ValueNotAllowedPatternCheck(),
            ValueTokenCheck(),
            ValueLengthCheck(max_len=config.max_url_cred_value_length),
            ValuePatternCheck()
        ]



from credsweeper.common.constants import Severity
from credsweeper.credentials.candidate import Candidate
from credsweeper.ml_model.features.feature import Feature


class RuleSeverity(Feature):
    """Categorical feature that corresponds to rule name."""

    def extract(self, candidate: Candidate) -> float:
        if Severity.CRITICAL == candidate.severity:
            return 1.0
        elif Severity.HIGH == candidate.severity:
            return 0.75
        elif Severity.MEDIUM == candidate.severity:
            return 0.5
        elif Severity.LOW == candidate.severity:
            return 0.25
        elif Severity.INFO == candidate.severity:
            return 0.0
        else:
            raise ValueError(f"Unknown type of severity: {candidate.severity}")



import logging
import struct
from abc import ABC
from typing import List, Optional, Generator, Tuple

from credsweeper.common.constants import LATIN_1, UTF_8
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider

logger = logging.getLogger(__name__)


class PngScanner(AbstractScanner, ABC):
    """Implements PNG scanning for text chunks"""

    @staticmethod
    def match(data: bytes) -> bool:
        """Returns True if prefix match"""
        if data.startswith(b"\x89PNG\r\n\x1a\n"):
            return True
        return False

    @staticmethod
    def yield_png_chunks(data: bytes) -> Generator[Tuple[int, str, bytes], None, None]:
        """Processes PNG chunks and yields offset, type and data"""
        offset = 8  # b"\x89PNG\r\n\x1a\n"
        data_limit = len(data) - 12
        while offset <= data_limit:
            chunk_size = struct.unpack(">I", data[offset:offset + 4])[0]
            chunk_type = data[offset + 4:offset + 8]
            offset += 8
            if len(data) < offset + chunk_size:
                raise ValueError(f"PNG chunk size {chunk_size} exceeds data limit 0x{offset:x}")
            match chunk_type:
                case b"IEND":
                    # https://www.w3.org/TR/png/#11IEND
                    break
                case b"tEXt":
                    # https://www.w3.org/TR/png/#11tEXt
                    keyword, text_data = data[offset:offset + chunk_size].split(b'\0', 1)
                    yield offset, f"PNG_TEXT:{keyword.decode(encoding=LATIN_1, errors='strict')}", text_data
                case b"zTXt":
                    # https://www.w3.org/TR/png/#11zTXt
                    keyword, ztxt_data = data[offset:offset + chunk_size].split(b'\0', 1)
                    if not ztxt_data.startswith(b'\0'):
                        raise ValueError(f"Unsupported compression method {ztxt_data[0]}")
                    yield offset, f"PNG_ZTXT:{keyword.decode(encoding=LATIN_1, errors='strict')}", ztxt_data[1:]
                case b"iTXt":
                    # https://www.w3.org/TR/png/#11iTXt
                    keyword, itxt_data = data[offset:offset + chunk_size].split(b'\0', 1)

                    if itxt_data.startswith(b"\x00\x00"):
                        compression = False
                    elif itxt_data.startswith(b"\x01\x00"):
                        compression = True
                    else:
                        raise ValueError(f"Unsupported compression {repr(itxt_data[:2])}")
                    lang_tag, itxt_data = itxt_data[2:].split(b'\0', 1)
                    trans_key, itxt_data = itxt_data.split(b'\0', 1)
                    yield (offset, f"PNG_ITXT_{'1' if compression else '0'}"
                           f":{keyword.decode(encoding=UTF_8)}"
                           f":{lang_tag.decode(encoding=UTF_8)}"
                           f":{trans_key.decode(encoding=UTF_8)}", itxt_data)
                case _:
                    pass
            # skip crc verification
            offset += chunk_size + 4

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan each row as structure with column name in key"""
        try:
            candidates: List[Candidate] = []
            for offset, chunk_type, data in PngScanner.yield_png_chunks(data_provider.data):
                png_content_provider = DataContentProvider(data=data,
                                                           file_path=data_provider.file_path,
                                                           file_type=data_provider.file_type,
                                                           info=f"{data_provider.info}|{chunk_type}:0x{offset:x}")
                new_limit = recursive_limit_size - len(data)
                png_candidates = self.recursive_scan(png_content_provider, depth, new_limit)
                candidates.extend(png_candidates)
            return candidates
        except Exception as exc:
            logger.error(exc)
        return None



from typing import Optional

from credsweeper.common.constants import MIN_VALUE_LENGTH, MAX_LINE_LENGTH
from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValueLengthCheck(Filter):
    """Check that candidate value length is between MIN and MAX."""

    def __init__(self,
                 config: Optional[Config] = None,
                 min_len: int = MIN_VALUE_LENGTH,
                 max_len: int = MAX_LINE_LENGTH) -> None:
        self.min_len = min_len
        self.max_len = max_len

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if self.min_len <= len(line_data.value) <= self.max_len:
            return False
        else:
            return True



from credsweeper.common.constants import GroupType
from credsweeper.config.config import Config
from credsweeper.filters import ValueLengthCheck, LineGitBinaryCheck
from credsweeper.filters import ValueSplitKeywordCheck
from credsweeper.filters.group.group import Group
from credsweeper.filters.line_uue_part_check import LineUUEPartCheck


class PasswordKeyword(Group):
    """PasswordKeyword"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.KEYWORD)
        self.filters.extend([
            ValueLengthCheck(max_len=config.max_password_value_length),
            ValueSplitKeywordCheck(),
            LineGitBinaryCheck(),
            LineUUEPartCheck()
        ])



import logging
import os.path
import sqlite3
import sys
import tempfile
from abc import ABC
from typing import List, Optional, Tuple, Any, Generator, Union

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider

logger = logging.getLogger(__name__)


class Sqlite3Scanner(AbstractScanner, ABC):
    """Implements SQLite3 database scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - SQLite Database"""
        if data.startswith(b"SQLite format 3\0"):
            return True
        return False

    @staticmethod
    def __walk(sqlite3db) -> Generator[Tuple[str, Any], None, None]:
        sqlite3db.row_factory = sqlite3.Row
        cursor = sqlite3db.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite_%';")
        for table in cursor.fetchall():
            table_name = table[0]
            try:
                cursor.execute(f"SELECT * FROM {table_name}")
                for row in cursor:
                    yield table_name, dict(row)
            except sqlite3.DatabaseError as exc:
                print(f"Error reading table {table_name}: {exc}")

    @staticmethod
    def walk_sqlite(data: bytes) -> Generator[Tuple[str, Any], None, None]:
        """Yields data from sqlite3 database"""
        if 10 < sys.version_info.minor:
            # Added in version 3.11
            with sqlite3.connect(":memory:") as sqlite3db:
                sqlite3db.deserialize(data)  # type: ignore
                yield from Sqlite3Scanner.__walk(sqlite3db)
        elif "nt" != os.name:
            # a tmpfile has to be used. TODO: remove when 3.10 will deprecate
            with tempfile.NamedTemporaryFile(suffix=".sqlite") as t:
                t.write(data)
                t.flush()
                with sqlite3.connect(t.name) as sqlite3db:
                    yield from Sqlite3Scanner.__walk(sqlite3db)
        elif "nt" == os.name:
            # windows trick. TODO: remove when 3.10 will deprecate
            with tempfile.NamedTemporaryFile(delete=False, suffix=".sqlite") as t:
                t.write(data)
                t.flush()
            sqlite3db = sqlite3.connect(t.name)
            yield from Sqlite3Scanner.__walk(sqlite3db)
            sqlite3db.close()
            if os.path.exists(t.name):
                os.remove(t.name)

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data file from .ar (debian) archive and launches data_scan"""
        try:
            candidates: List[Candidate] = []
            new_limit = recursive_limit_size - len(data_provider.data)
            for table, row in self.walk_sqlite(data_provider.data):
                struct_content_provider = StructContentProvider(struct=row,
                                                                file_path=data_provider.file_path,
                                                                file_type=data_provider.file_type,
                                                                info=f"{data_provider.info}|SQLite3.{table}")
                if new_candidates := self.structure_scan(struct_content_provider, depth, new_limit):
                    candidates.extend(new_candidates)
            return candidates
        except Exception as exc:
            logger.error(exc)
        return None



import math
from functools import cache

from credsweeper.filters.value_entropy_base_check import ValueEntropyBaseCheck


class ValueEntropyBase32Check(ValueEntropyBaseCheck):
    """Base32 entropy check"""

    @staticmethod
    @cache
    def get_min_data_entropy(x: int) -> float:
        """Returns average entropy for size of random data. Precalculated data is applied for speedup"""
        if 8 <= x < 17:
            y = 0.80569236 * math.log2(x) + 0.13439734
        elif 17 <= x < 33:
            y = 0.66350481 * math.log2(x) + 0.71143862
        elif 33 <= x:
            y = 4.04
        else:
            y = 0
        return y



import logging
from abc import ABC
from typing import List, Optional, Union

from lxml import etree

from credsweeper.common.constants import MIN_DATA_LEN, MAX_LINE_LENGTH
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class TmxScanner(AbstractScanner, ABC):
    """Realises tmX files scanning for values only. Image tags are skipped."""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """Used to detect tm7,tm6,etc. (ThreadModeling) format."""
        for opening_tag, closing_tag in [(b"<ThreatModel", b"</ThreatModel>"),
                                         (b"<KnowledgeBase", b"</KnowledgeBase>")]:
            opening_pos = data.find(opening_tag, 0, MAX_LINE_LENGTH)
            if 0 <= opening_pos < data.find(closing_tag, opening_pos):
                # opening and closing tags were found - suppose it is an HTML
                return True
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to represent data as xml text and scan as text lines"""
        try:
            lines = []
            # the format is always in single line xlm, so line numbers are not actual
            tree = etree.fromstring(data_provider.data)
            for element in tree.iter():
                tag = Util.extract_element_data(element, "tag")
                if "Image" in tag:
                    continue
                text = Util.extract_element_data(element, "text")
                if MIN_DATA_LEN > len(text):
                    continue
                lines.append(text)
            tmx_data_provider = StringContentProvider(lines=lines,
                                                      file_path=data_provider.file_path,
                                                      file_type=data_provider.file_type,
                                                      info=f"{data_provider.info}|TMX")
            return self.scanner.scan(tmx_data_provider)
        except Exception as exc:
            logger.warning("Cannot processed tmX file %s %s", str(data_provider.file_path), str(exc))
        return None



import io
import logging
from abc import ABC
from typing import List, Optional, Union

import rpmfile

from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.utils.util import Util

logger = logging.getLogger(__name__)


class RpmScanner(AbstractScanner, ABC):
    """Implements rpm scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures"""
        if data.startswith(b"\xED\xAB\xEE\xDB"):
            return True
        return False

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts files one by one from the package type and launches recursive scan"""
        try:
            candidates = []
            with rpmfile.open(fileobj=io.BytesIO(data_provider.data)) as rpm_file:
                for member in rpm_file.getmembers():
                    # skip directory
                    if 0 != member.isdir:
                        continue
                    if FilePathExtractor.check_exclude_file(self.config, member.name):
                        continue
                    if 0 > recursive_limit_size - member.size:
                        logger.error(f"{member.filename}: size {member.size}"
                                     f" is over limit {recursive_limit_size} depth:{depth}")
                        continue
                    rpm_content_provider = DataContentProvider(data=rpm_file.extractfile(member).read(),
                                                               file_path=data_provider.file_path,
                                                               file_type=Util.get_extension(member.name),
                                                               info=f"{data_provider.info}|RPM:{member.name}")
                    new_limit = recursive_limit_size - len(rpm_content_provider.data)
                    rpm_candidates = self.recursive_scan(rpm_content_provider, depth, new_limit)
                    candidates.extend(rpm_candidates)
            return candidates
        except Exception as rpm_exc:
            logger.error(f"{data_provider.file_path}:{rpm_exc}")
        return None



import io
import logging
import struct
from abc import ABC
from typing import List, Optional, Union

from credsweeper.common.constants import UTF_8
from credsweeper.credentials.candidate import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.struct_content_provider import StructContentProvider

logger = logging.getLogger(__name__)


class JclassScanner(AbstractScanner, ABC):
    """Implements java .class scanning"""

    @staticmethod
    def match(data: Union[bytes, bytearray]) -> bool:
        """According https://en.wikipedia.org/wiki/List_of_file_signatures - java class"""
        if data.startswith(b"\xCA\xFE\xBA\xBE"):
            return True
        return False

    @staticmethod
    def u2(stream: io.BytesIO) -> int:
        """Extracts unsigned 16 bit big-endian"""
        return int(struct.unpack(">H", stream.read(2))[0])

    @staticmethod
    def get_utf8_constants(stream: io.BytesIO) -> List[str]:
        """Extracts only Utf8 constants from java ClassFile"""
        result = []
        # actual number of items is one less!
        items_counter = JclassScanner.u2(stream) - 1
        while 0 < items_counter:
            items_counter -= 1
            # uint8
            tag = int(stream.read(1)[0])
            if 1 == tag:
                # UTF-8 string in bytes may be bigger than in characters
                length = JclassScanner.u2(stream)
                data = stream.read(int(length))
                value = data.decode(encoding=UTF_8, errors="replace")
                result.append(value)
            elif tag in (3, 4, 9, 10, 11, 12, 18):
                _ = stream.read(4)
            elif tag in (7, 8, 16):
                _ = stream.read(2)
            elif tag in (5, 6):
                _ = stream.read(8)
                # long and double types use two indexes
                items_counter -= 1
            elif 15 == tag:
                _ = stream.read(3)
            else:
                logger.error(f"Unknown tag {tag}")
                break
        return result

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts data from binary"""
        try:
            stream = io.BytesIO(data_provider.data)
            stream.read(4)  # magic
            minor = JclassScanner.u2(stream)
            major = JclassScanner.u2(stream)
            constants = JclassScanner.get_utf8_constants(stream)
            struct_content_provider = StructContentProvider(struct=constants,
                                                            file_path=data_provider.file_path,
                                                            file_type=data_provider.file_type,
                                                            info=f"{data_provider.info}|Java.{major}.{minor}")
            new_limit = recursive_limit_size - sum(len(x) for x in constants)
            candidates = self.structure_scan(struct_content_provider, depth, new_limit)
            return candidates
        except Exception as jclass_exc:
            logger.error(f"{data_provider.file_path}:{jclass_exc}")
        return None



import re
from typing import Optional

from credsweeper.common.constants import DEFAULT_PATTERN_LEN, MAX_LINE_LENGTH, MIN_DATA_LEN
from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter


class ValuePatternCheck(Filter):
    """Check if candidate value contain specific pattern.

    Similar to linguistic sequences of characters, random strings shouldn't contain math sequences of
    characters. Based on "How Bad Can It Git? Characterizing Secret Leakage in Public GitHub Repositories", details:
    https://www.ndss-symposium.org/ndss-paper/how-bad-can-it-git-characterizing-secret-leakage-in-public-github-repositories/
    PatternCheck checks the occurrence in "line_data.value" of three types of sequence:

    - N or more identical characters in sequence, example: "AAAA", "1111" ...
    - N or more increasing characters sequentially, example: "abcd", "1234" ...
    - N or more decreasing characters sequentially, example: "dcba", "4321" ...

    Default pattern LEN is 4
    """

    MAX_PATTERN_LENGTH = int(MAX_LINE_LENGTH).bit_length()

    def __init__(self, config: Optional[Config] = None, pattern_len: Optional[int] = None):
        """Create ValuePatternCheck with a specific pattern_len to check.

        Args:
            config: pattern len to use during check. DEFAULT_PATTERN_LEN by default
            pattern_len: size of constant pattern length for any value size or None for dynamic pattern size

        """
        patterns_count = 1 + ValuePatternCheck.MAX_PATTERN_LENGTH
        if pattern_len is None:
            self.pattern_len = -1
            # pattern length depends on value length
            self.pattern_lengths = [max(x, DEFAULT_PATTERN_LEN) for x in range(patterns_count)]
            self.patterns = [ValuePatternCheck.get_pattern(x) for x in range(patterns_count)]
        elif isinstance(pattern_len, int) and DEFAULT_PATTERN_LEN <= pattern_len:
            self.pattern_len = pattern_len
            # constant pattern for any value length
            self.pattern_lengths = [pattern_len] * patterns_count
            self.patterns = [ValuePatternCheck.get_pattern(pattern_len)] * patterns_count
        else:
            raise ValueError(f"Wrong type of pattern length {type(pattern_len)} = {repr(pattern_len)}")

    @staticmethod
    def get_pattern(pattern_len: int) -> re.Pattern:
        """Creates regex pattern to find N or more identical characters in sequence"""
        pattern_length = max(DEFAULT_PATTERN_LEN, pattern_len)
        if MIN_DATA_LEN <= pattern_length:
            # base64 long sequences may contain 0x00 or 0xFF inside
            pattern = fr"([^\sA/_])\1{{{str(pattern_length-1)},}}"
        else:
            # up to 256 symbols length
            pattern = fr"(\S)\1{{{str(pattern_length-1)},}}"
        return re.compile(pattern)

    def equal_pattern_check(self, value: str, bit_length: int) -> bool:
        """Check if candidate value contain 4 and more same chars or numbers sequences.

        Args:
            value: string variable, credential candidate value
            bit_length: speedup for len(value).bit_length()

        Return:
            True if contain and False if not

        """
        if self.patterns[bit_length].search(value):
            return True
        return False

    def ascending_pattern_check(self, value: str, bit_length: int) -> bool:
        """Check if candidate value contain 4 and more ascending chars or numbers sequences.

        Arg:
            value: credential candidate value
            bit_length: speedup for len(value).bit_length()

        Return:
            True if contain and False if not

        """
        count = 1
        for key in range(len(value) - 1):
            if ord(value[key + 1]) - ord(value[key]) == 1:
                count += 1
            else:
                count = 1
                continue
            if count == self.pattern_lengths[bit_length]:
                return True
        return False

    def descending_pattern_check(self, value: str, bit_length: int) -> bool:
        """Check if candidate value contain 4 and more descending chars or numbers sequences.

        Arg:
            value: string variable, credential candidate value
            bit_length: speedup for len(value).bit_length()

        Return:
            boolean variable. True if contain and False if not

        """
        count = 1
        for key in range(len(value) - 1):
            if ord(value[key]) - ord(value[key + 1]) == 1:
                count += 1
            else:
                count = 1
                continue
            if count == self.pattern_lengths[bit_length]:
                return True
        return False

    def check_val(self, value: str, bit_length: int) -> bool:
        """Cumulative value check.

        Arg:
            value: string variable, credential candidate value
            bit_length: speedup for len(value).bit_length()

        Return:
            boolean variable. True if contain and False if not

        """
        if self.equal_pattern_check(value, bit_length):
            return True
        if self.ascending_pattern_check(value, bit_length):
            return True
        if self.descending_pattern_check(value, bit_length):
            return True
        return False

    def duple_pattern_check(self, value: str, bit_length: int) -> bool:
        """Check if candidate value is a duplet value with possible patterns.

        Arg:
            value: string variable, credential candidate value
            bit_length: speedup for len(value).bit_length()

        Return:
            boolean variable. True if contain and False if not

        """
        even_value = value[0::2]
        if self.check_val(even_value, bit_length):
            odd_value = value[1::2]
            if self.check_val(odd_value, bit_length):
                return True
        return False

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Arg:
            line_data: LineData object, credential candidate data
            target: multiline target from which line data was obtained

        Return:
            boolean variable. True, if need to filter candidate and False if left

        """
        value_length = len(line_data.value)
        bit_length = max(DEFAULT_PATTERN_LEN, value_length.bit_length())

        if ValuePatternCheck.MAX_PATTERN_LENGTH < bit_length:
            # huge values may contain anything
            return False

        if 0 <= value_length < self.pattern_len or value_length < self.pattern_lengths[bit_length]:
            # too short value
            return True

        if self.check_val(line_data.value, bit_length):
            return True

        if 2 * self.pattern_lengths[bit_length] <= value_length \
                and self.duple_pattern_check(line_data.value, bit_length):
            return True

        return False



from credsweeper.app import CredSweeper
from credsweeper.common.constants import ThresholdPreset, Severity, Confidence
from credsweeper.file_handler.byte_content_provider import ByteContentProvider
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider
from credsweeper.file_handler.text_content_provider import TextContentProvider

from credsweeper.ml_model.ml_validator import MlValidator

__all__ = [
    "ByteContentProvider",  #
    "Confidence",  #
    "ContentProvider",  #
    "CredSweeper",  #
    "DataContentProvider",  #
    "DiffContentProvider",  #
    "MlValidator",  #
    "Severity",  #
    "StringContentProvider",  #
    "TextContentProvider",  #
    "ThresholdPreset",  #
    "__version__"
]

__version__ = "1.15.1"



import re
from typing import Optional

from credsweeper.config.config import Config
from credsweeper.credentials.line_data import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters.filter import Filter
from credsweeper.utils.util import Util


class ValueAllowlistCheck(Filter):
    """Check that the patterns do not MATCH the candidate value."""

    ALLOWED = [
        r"ENC\(.*\)",  #
        r"ENC\[.*\]",  #
        r"\$\{(\*|[0-9]+|[a-z_].*)\}",  #
        r"\$[0-9]+(\s|$)",  #
        r"\$\$[a-z_]+(\^%[0-9a-z_]+)?",  #
        r"#\{.+\}",  # Ruby: String Interpolation
        r"\{\{.+\}\}",  #
        r".*@@@hl@@@(암호|비번|PW|PASS)@@@endhl@@@",  #
    ]

    ALLOWED_PATTERN = re.compile(Util.get_regex_combine_or(ALLOWED), flags=re.IGNORECASE)

    ALLOWED_QUOTED = [
        r"\$[a-z_][0-9a-z_]+((::|->|\.)[a-z_]|\[|$)",  #
        r"\$\([^)]+\)",  #
        r".*\*\*\*",  #
    ]

    ALLOWED_QUOTED_PATTERN = re.compile(Util.get_regex_combine_or(ALLOWED_QUOTED), flags=re.IGNORECASE)

    ALLOWED_UNQUOTED = [
        r"[~a-z0-9_]+((\.|->)[a-z0-9_]+)+\(.*$",  #
        r"\$[a-z_][0-9a-z_]+((::|->|\.)[a-z_]|\[|$)",  #
        r"\$\([.0-9a-z_-]+",  #
        r".*\*\*\*\*\*",  #
    ]

    ALLOWED_UNQUOTED_PATTERN = re.compile(Util.get_regex_combine_or(ALLOWED_UNQUOTED), flags=re.IGNORECASE)

    def __init__(self, config: Optional[Config] = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.is_well_quoted_value:
            if self.ALLOWED_PATTERN.match(line_data.value) or self.ALLOWED_QUOTED_PATTERN.match(line_data.value):
                return True
        else:
            value = line_data.wrap + line_data.value if line_data.wrap else line_data.value
            if self.ALLOWED_PATTERN.match(value) or self.ALLOWED_UNQUOTED_PATTERN.match(value):
                return True
        return False
